##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Classes: PluggableAuthService
"""
import logging

from AccessControl import ClassSecurityInfo, ModuleSecurityInfo
from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManagement import getSecurityManager
from AccessControl.SecurityManagement import noSecurityManager
from AccessControl.Permissions import manage_users as ManageUsers
from AccessControl.User import nobody
from AccessControl.SpecialUsers import emergency_user
from Acquisition import Implicit
from Acquisition import aq_parent
from Acquisition import aq_base
from Acquisition import aq_inner
from App.class_init import default__class_init__ as InitializeClass
from App.ImageFile import ImageFile
from OFS.Folder import Folder
from OFS.Cache import Cacheable
from OFS.interfaces import IObjectManager
from OFS.interfaces import IPropertyManager
from Products.StandardCacheManagers.RAMCacheManager import RAMCacheManager
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from zExceptions import Unauthorized
from ZPublisher import BeforeTraverse
from ZTUtils import Batch
from zope.event import notify

from Products.PluginRegistry.PluginRegistry import PluginRegistry

from .events import PrincipalCreated
from .interfaces.authservice import IPluggableAuthService
from .interfaces.authservice import _noroles
from .interfaces.plugins import IExtractionPlugin
from .interfaces.plugins import ILoginPasswordHostExtractionPlugin
from .interfaces.plugins import IAuthenticationPlugin
from .interfaces.plugins import IChallengePlugin
from .interfaces.plugins import ICredentialsUpdatePlugin
from .interfaces.plugins import ICredentialsResetPlugin
from .interfaces.plugins import IUserFactoryPlugin
from .interfaces.plugins import IAnonymousUserFactoryPlugin
from .interfaces.plugins import IPropertiesPlugin
from .interfaces.plugins import IGroupsPlugin
from .interfaces.plugins import IRolesPlugin
from .interfaces.plugins import IUpdatePlugin
from .interfaces.plugins import IValidationPlugin
from .interfaces.plugins import IUserEnumerationPlugin
from .interfaces.plugins import IUserAdderPlugin
from .interfaces.plugins import IGroupEnumerationPlugin
from .interfaces.plugins import IRoleEnumerationPlugin
from .interfaces.plugins import IRoleAssignerPlugin
from .interfaces.plugins import INotCompetentPlugin
from .interfaces.plugins import IChallengeProtocolChooser
from .interfaces.plugins import IRequestTypeSniffer
from .permissions import SearchPrincipals
from .PropertiedUser import PropertiedUser
from .utils import _wwwdir
from .utils import createViewName
from .utils import createKeywords
from .utils import classImplements

security = ModuleSecurityInfo(
    'Products.PluggableAuthService.PluggableAuthService' )

logger = logging.getLogger('PluggableAuthService')

#   Errors which plugins may raise, and which we suppress:
_SWALLOWABLE_PLUGIN_EXCEPTIONS = ( NameError
                                 , AttributeError
                                 , KeyError
                                 , TypeError
                                 , ValueError
                                 )
# except if they tell us not to do so
def reraise(plugin):
    try:
        doreraise = plugin._dont_swallow_my_exceptions
    except AttributeError:
        return
    if doreraise:
        raise

MultiPlugins = []
def registerMultiPlugin(meta_type):
    """ Register a 'multi-plugin' in order to expose it to the Add List
    """
    if meta_type in MultiPlugins:
        raise RuntimeError('Meta-type (%s) already available to Add List'
                           % meta_type)
    MultiPlugins.append(meta_type)


class DumbHTTPExtractor( Implicit ):

    security = ClassSecurityInfo()

    security.declarePrivate( 'extractCredentials' )
    def extractCredentials( self, request ):

        """ Pull HTTP credentials out of the request.
        """
        creds = {}
        login_pw = request._authUserPW()

        if login_pw is not None:
            name, password = login_pw

            creds[ 'login' ] = name
            creds[ 'password' ] = password
            creds[ 'remote_host' ] = request.get( 'REMOTE_HOST', '' )

            try:
                creds[ 'remote_address' ] = request.getClientAddr()
            except AttributeError:
                creds[ 'remote_address' ] = request.get( 'REMOTE_ADDR', '' )

        return creds

classImplements( DumbHTTPExtractor
               , ILoginPasswordHostExtractionPlugin
               )

InitializeClass( DumbHTTPExtractor )


class EmergencyUserAuthenticator( Implicit ):

    security = ClassSecurityInfo()

    security.declarePrivate( 'authenticateCredentials' )
    def authenticateCredentials( self, credentials ):

        """ Check credentials against the emergency user.
        """
        if isinstance( credentials, dict ):

            eu = emergency_user
            eu_name = eu.getUserName()
            login = credentials.get( 'login' )

            if login == eu_name:
                password = credentials.get( 'password' )

                if eu.authenticate( password, {} ):
                    return (eu_name, None)

        return (None, None)

classImplements( EmergencyUserAuthenticator
               , IAuthenticationPlugin
               )

InitializeClass( EmergencyUserAuthenticator )


class PluggableAuthService( Folder, Cacheable ):

    """ All-singing, all-dancing user folder.
    """
    security = ClassSecurityInfo()

    meta_type = 'Pluggable Auth Service'

    _id = id = 'acl_users'

    _emergency_user = emergency_user
    _nobody = nobody

    maxlistusers = -1   # Don't allow local role form to try to list us!

    # Method for transforming a login name.  This needs to be the name
    # of a method on this plugin.  See the applyTransform method.
    login_transform = ''

    _properties = (
        dict(id='title', type='string', mode='w',
             label='Title'),
        dict(id='login_transform', type='string', mode='w',
             label='Transform to apply to login name'),
             )


    def getId( self ):

        return self._id

    #
    #   IUserFolder implementation
    #
    security.declareProtected( ManageUsers, 'getUser' )
    def getUser( self, name ):

        """ See IUserFolder.
        """
        plugins = self._getOb( 'plugins' )

        name = self.applyTransform( name )
        user_info = self._verifyUser( plugins, login=name )

        if not user_info:
            return None

        return self._findUser( plugins, user_info['id'], user_info['login'])

    security.declareProtected( ManageUsers, 'getUserById' )
    def getUserById( self, id, default=None ):

        """ See IUserFolder.
        """
        plugins = self._getOb( 'plugins' )

        user_info = self._verifyUser( plugins, user_id=id )

        if not user_info:
            return default

        return self._findUser( plugins, user_info['id'], user_info['login'])

    security.declarePublic( 'validate' )     # XXX: public?
    def validate( self, request, auth='', roles=_noroles ):

        """ See IUserFolder.
        """
        plugins = self._getOb( 'plugins' )
        is_top = self._isTop()

        if not is_top and self._isNotCompetent( request, plugins ):
            # this user folder should not try to authenticate this request
            return None

        user_ids = self._extractUserIds(request, plugins)
        ( accessed
        , container
        , name
        , value
        ) = self._getObjectContext( request[ 'PUBLISHED' ], request )

        for user_id, login in user_ids:

            user = self._findUser(plugins, user_id, login, request=request)

            if aq_base( user ) is emergency_user:

                if is_top:
                    return user
                else:
                    return None

            if self._authorizeUser( user
                                  , accessed
                                  , container
                                  , name
                                  , value
                                  , roles
                                  ):
                return user

        if not is_top:
            return None

        #
        #   No other user folder above us can satisfy, and we have no user;
        #   return a constructed anonymous only if anonymous is authorized.
        #
        anonymous = self._createAnonymousUser( plugins )
        if self._authorizeUser( anonymous
                              , accessed
                              , container
                              , name
                              , value
                              , roles
                              ):
            return anonymous

        return None

    security.declareProtected( SearchPrincipals, 'searchUsers')
    def searchUsers(self, **kw):
        """ Search for users
        """
        search_id = kw.get( 'id', None )
        search_name = kw.get( 'name', None )

        result = []
        max_results = kw.get('max_results', '')
        sort_by = kw.get('sort_by', '')

        # We apply sorting and slicing here across all sets, so don't
        # make the plugin do it
        if sort_by:
            del kw['sort_by']
        if max_results:
            del kw['max_results']
        if search_name:
            if kw.get('id') is not None:
                del kw['id'] # don't even bother searching by id
            # Note: name can be a sequence.
            kw['login'] = self.applyTransform( kw['name'] )
        if kw.get('login', None):
            kw['login'] = self.applyTransform( kw['login'] )

        plugins = self._getOb( 'plugins' )
        enumerators = plugins.listPlugins( IUserEnumerationPlugin )

        for enumerator_id, enum in enumerators:
            try:
                user_list = enum.enumerateUsers(**kw)
                for user_info in user_list:
                    info = {}
                    info.update( user_info )
                    info[ 'userid' ] = info[ 'id' ]
                    info[ 'principal_type' ] = 'user'
                    if not info.has_key('title'):
                        info[ 'title' ] = info[ 'login' ]
                    result.append(info)

            except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                reraise(enum)
                logger.debug( 'UserEnumerationPlugin %s error' % enumerator_id
                            , exc_info=True
                            )

        if sort_by:
            result.sort( lambda a, b: cmp( a.get(sort_by, '').lower()
                                         , b.get(sort_by, '').lower()
                                         ) )

        if max_results:
            try:
                max_results = int(max_results)
                result = result[:max_results]
            except ValueError:
                pass

        return tuple(result)

    security.declareProtected( SearchPrincipals, 'searchGroups')
    def searchGroups(self, **kw):
        """ Search for groups
        """
        search_id = kw.get( 'id', None )
        search_name = kw.get( 'name', None )

        result = []
        max_results = kw.get('max_results', '')
        sort_by = kw.get('sort_by', '')

        # We apply sorting and slicing here across all sets, so don't
        # make the plugin do it
        if sort_by:
            del kw['sort_by']
        if max_results:
            del kw['max_results']
        if search_name:
            if kw.get('id') is not None:
                del kw['id']
            if not kw.has_key('title'):
                kw['title'] = kw['name']

        plugins = self._getOb( 'plugins' )
        enumerators = plugins.listPlugins( IGroupEnumerationPlugin )

        for enumerator_id, enum in enumerators:
            try:
                 group_list = enum.enumerateGroups(**kw)
                 for group_info in group_list:
                    info = {}
                    info.update( group_info )
                    info[ 'groupid' ] = info[ 'id' ]
                    info[ 'principal_type' ] = 'group'
                    if not info.has_key('title'):
                        info[ 'title' ] = "(Group) %s" % info[ 'groupid' ]
                    result.append(info)
            except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                reraise(enum)
                logger.debug( 'GroupEnumerationPlugin %s error' % enumerator_id
                            , exc_info=True
                            )

        if sort_by:
            result.sort( lambda a, b: cmp( a.get(sort_by, '').lower()
                                         , b.get(sort_by, '').lower()
                                         ) )

        if max_results:
            try:
                max_results = int(max_results)
                result = result[:max_results + 1]
            except ValueError:
                pass

        return tuple(result)

    security.declareProtected( SearchPrincipals, 'searchPrincipals')
    def searchPrincipals(self, groups_first=False, **kw):
        """ Search for principals (users, groups, or both)
        """
        max_results = kw.get( 'max_results', '' )

        search_id = kw.get( 'id', None )
        search_name = kw.get( 'name', None )
        if search_name:
            if not kw.has_key('title'):
                kw['title'] = search_name
            kw['login'] = search_name
            
        # For groups we search the original name
        # (e.g. Administrators), for users we apply the transform,
        # which could lowercase the name.
        groups = [ d.copy() for d in self.searchGroups( **kw ) ]
        if kw.get('login', None):
            kw['login'] = self.applyTransform( kw['login'] )
        users = [ d.copy() for d in self.searchUsers( **kw ) ]

        if groups_first:
            result = groups + users
        else:
            result = users + groups

        if max_results:
            try:
                max_results = int( max_results )
                result = result[ :max_results + 1 ]
            except ValueError:
                pass

        return tuple( result )

    security.declarePrivate( '__creatable_by_emergency_user__' )
    def __creatable_by_emergency_user__( self ):
        return 1

    security.declarePrivate( '_setObject' )
    def _setObject( self, id, object, roles=None, user=None, set_owner=0 ):
        #
        #   Override ObjectManager's version to change the default for
        #   'set_owner' (we don't want to enforce ownership on contained
        #   objects).
        Folder._setObject( self, id, object, roles, user, set_owner )

    security.declarePrivate( '_delOb' )
    def _delOb( self, id ):
        #
        #   Override ObjectManager's version to clean up any plugin
        #   registrations for the deleted object
        #
        plugins = self._getOb( 'plugins', None )

        if plugins is not None:
            plugins.removePluginById( id )

        Folder._delOb( self, id )

    #
    # ZMI stuff
    #
    arrow_right_gif = ImageFile( 'www/arrow-right.gif', globals() )
    arrow_left_gif = ImageFile( 'www/arrow-left.gif', globals() )
    arrow_up_gif = ImageFile( 'www/arrow-up.gif', globals() )
    arrow_down_gif = ImageFile( 'www/arrow-down.gif', globals() )

    security.declareProtected(ManageUsers, 'manage_search')
    manage_search = PageTemplateFile('www/pasSearch', globals())

    manage_options = ( Folder.manage_options[:1]
                      + ( { 'label' : 'Search'
                          , 'action': 'manage_search' }
                        ,
                        )
                      + Folder.manage_options[2:]
                      + Cacheable.manage_options
                      )

    security.declareProtected(ManageUsers, 'resultsBatch')
    def resultsBatch(self, results, REQUEST, size=20, orphan=2, overlap=0):
        """ ZMI helper for getting batching for displaying search results
        """
        try:
            start_val = REQUEST.get('batch_start', '0')
            start = int(start_val)
            size = int(REQUEST.get('batch_size', size))
        except ValueError:
            start = 0

        batch = Batch(results, size, start, 0, orphan, overlap)

        if batch.end < len(results):
            qs = self._getBatchLink( REQUEST.get('QUERY_STRING', '')
                                   , start
                                   , batch.end
                                   )
            REQUEST.set( 'next_batch_url'
                       , '%s?%s' % (REQUEST.get('URL'), qs)
                       )

        if start > 0:
            new_start = start - size - 1

            if new_start < 0:
                new_start = 0

            qs = self._getBatchLink( REQUEST.get('QUERY_STRING', '')
                                   , start
                                   , new_start
                                   )
            REQUEST.set( 'previous_batch_url'
                       , '%s?%s' % (REQUEST.get('URL'), qs)
                       )

        return batch


    security.declarePrivate('_getBatchLink')
    def _getBatchLink(self, qs, old_start, new_start):
        """ Internal helper to generate correct query strings
        """
        if new_start is not None:
            if not qs:
                qs = 'batch_start=%d' % new_start
            elif qs.startswith('batch_start='):
                qs = qs.replace( 'batch_start=%d' % old_start
                               , 'batch_start=%d' % new_start
                               )
            elif qs.find('&batch_start=') != -1:
                qs = qs.replace( '&batch_start=%d' % old_start
                               , '&batch_start=%d' % new_start
                               )
            else:
                qs = '%s&batch_start=%d' % (qs, new_start)

        return qs


    #
    #   Helper methods
    #
    security.declarePrivate( '_isNotCompetent' )
    def _isNotCompetent( self, request, plugins ):

        """ return true when this user folder should not try authentication.

        Never called for top level user folder.
        """
        try:
            not_competents = plugins.listPlugins( INotCompetentPlugin )
        except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
            logger.debug('NotCompetent plugin listing error', exc_info=True)
            not_competents = ()

        for not_competent_id, not_competent in not_competents:
            try:
                if not_competent.isNotCompetentToAuthenticate(request):
                    return True
            except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                reraise(not_competent)
                logger.debug( 'NotCompetentPlugin %s error' % not_competent_id
                            , exc_info=True
                            )
                continue
        return False

        
    security.declarePrivate( '_extractUserIds' )
    def _extractUserIds( self, request, plugins ):

        """ request -> [ validated_user_id ]

        o For each set of extracted credentials, try to authenticate
          a user;  accumulate a list of the IDs of such users over all
          our authentication and extraction plugins.
        """
        try:
            extractors = plugins.listPlugins( IExtractionPlugin )
        except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
            logger.debug('Extractor plugin listing error', exc_info=True)
            extractors = ()

        if not extractors:
            extractors = ( ( 'default', DumbHTTPExtractor() ), )

        try:
            authenticators = plugins.listPlugins( IAuthenticationPlugin )
        except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
            logger.debug('Authenticator plugin listing error', exc_info=True)
            authenticators = ()

        result = []

        for extractor_id, extractor in extractors:

            try:
                credentials = extractor.extractCredentials( request )
            except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                reraise(extractor)
                logger.debug( 'ExtractionPlugin %s error' % extractor_id
                            , exc_info=True
                            )
                continue

            if credentials:

                try:
                    credentials[ 'extractor' ] = extractor_id # XXX: in key?
                    # Test if ObjectCacheEntries.aggregateIndex would work
                    items = credentials.items()
                    items.sort()
                except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                    # XXX: would reraise be good here, and which plugin to ask
                    # whether not to swallow the exception - the extractor?
                    logger.debug( 'Credentials error: %s' % credentials
                                , exc_info=True
                                )
                    continue

                # First try to authenticate against the emergency
                # user and return immediately if authenticated
                user_id, name = self._tryEmergencyUserAuthentication(
                                                            credentials )

                if user_id is not None:
                    return [ ( user_id, name ) ]

                # Now see if the user ids can be retrieved from the cache
                credentials['login'] = self.applyTransform( credentials.get('login') )
                view_name = createViewName('_extractUserIds',
                                           credentials.get('login'))
                keywords = createKeywords(**credentials)
                user_ids = self.ZCacheable_get( view_name=view_name
                                              , keywords=keywords
                                              , default=None
                                              )
                if user_ids is None:
                    user_ids = []

                    for authenticator_id, auth in authenticators:

                        try:
                            uid_and_info = auth.authenticateCredentials(
                                credentials )

                            if uid_and_info is None:
                                continue

                            user_id, info = uid_and_info

                        except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                            reraise(auth)
                            msg = 'AuthenticationPlugin %s error' % (
                                    authenticator_id, )
                            logger.debug(msg, exc_info=True)
                            continue

                        if user_id is not None:
                            user_ids.append( (user_id, info) )

                    if user_ids:
                        self.ZCacheable_set( user_ids
                                           , view_name=view_name
                                           , keywords=keywords
                                           )

                result.extend( user_ids )

        # Emergency user via HTTP basic auth always wins
        user_id, name = self._tryEmergencyUserAuthentication(
                DumbHTTPExtractor().extractCredentials( request ) )

        if user_id is not None:
            return [ ( user_id, name ) ]

        return result

    security.declarePrivate( '_tryEmergencyUserAuthentication' )
    def _tryEmergencyUserAuthentication( self, credentials ):

        """ credentials -> emergency_user or None
        """
        try:
            eua = EmergencyUserAuthenticator()
            user_id, name = eua.authenticateCredentials( credentials )
        except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
            logger.debug('Credentials error: %s' % credentials, exc_info=True)
            user_id, name = ( None, None )

        return ( user_id, name )

    security.declarePrivate( '_getGroupsForPrincipal' )
    def _getGroupsForPrincipal( self
                              , principal
                              , request=None
                              , plugins=None
                              , ignore_plugins=None
                              ):
        all_groups = {}

        if ignore_plugins is None:
            ignore_plugins = ()

        if plugins is None:
            plugins = self._getOb( 'plugins' )
        groupmakers = plugins.listPlugins( IGroupsPlugin )

        for groupmaker_id, groupmaker in groupmakers:

            if groupmaker_id in ignore_plugins:
                continue
            groups = groupmaker.getGroupsForPrincipal( principal, request )
            for group in groups:
                principal._addGroups( [ group ] )
                all_groups[ group ] = 1

        return all_groups.keys()

    security.declarePrivate( '_createAnonymousUser' )
    def _createAnonymousUser( self, plugins ):

        """ Allow IAnonymousUserFactoryPlugins to create or fall back.
        """
        factories = plugins.listPlugins( IAnonymousUserFactoryPlugin )

        for factory_id, factory in factories:

            anon = factory.createAnonymousUser()

            if anon is not None:
                return anon.__of__( self )

        return nobody.__of__( self )

    security.declarePrivate( '_createUser' )
    def _createUser( self, plugins, user_id, name ):

        """ Allow IUserFactoryPlugins to create, or fall back to default.
        """
        name = self.applyTransform( name )
        factories = plugins.listPlugins( IUserFactoryPlugin )

        for factory_id, factory in factories:

            user = factory.createUser( user_id, name )

            if user is not None:
                return user.__of__( self )

        return PropertiedUser( user_id, name ).__of__( self )

    security.declarePrivate( '_findUser' )
    def _findUser( self, plugins, user_id, name=None, request=None ):

        """ user_id -> decorated_user
        """
        if user_id == self._emergency_user.getUserName():
            return self._emergency_user

        # See if the user can be retrieved from the cache
        view_name = createViewName('_findUser', user_id)
        name = self.applyTransform( name )
        keywords = createKeywords(user_id=user_id, name=name)
        user = self.ZCacheable_get( view_name=view_name
                                  , keywords=keywords
                                  , default=None
                                  )

        if user is None:

            user = self._createUser( plugins, user_id, name )
            propfinders = plugins.listPlugins( IPropertiesPlugin )

            for propfinder_id, propfinder in propfinders:

                data = propfinder.getPropertiesForUser( user, request )
                if data:
                    user.addPropertysheet( propfinder_id, data )

            groups = self._getGroupsForPrincipal( user, request
                                                , plugins=plugins )
            user._addGroups( groups )

            rolemakers = plugins.listPlugins( IRolesPlugin )

            for rolemaker_id, rolemaker in rolemakers:
                try:
                    roles = rolemaker.getRolesForPrincipal( user, request )
                except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                    logger.debug( 'IRolesPlugin %s error' % rolemaker_id
                                , exc_info=True
                                )
                else:
                    if roles:
                        user._addRoles( roles )

            user._addRoles( ['Authenticated'] )

            # Cache the user if caching is enabled
            base_user = aq_base(user)
            if getattr(base_user, '_p_jar', None) is None:
                self.ZCacheable_set( base_user
                                   , view_name=view_name
                                   , keywords=keywords
                                   )

        return user.__of__( self )

    security.declarePrivate( '_verifyUser' )
    def _verifyUser( self, plugins, user_id=None, login=None ):
        """ user_id -> info_dict or None
        """
        if user_id is None and login is None:
            # Avoid possible hugely expensive and/or wrong behavior of
            # plugin enumerators.
            return None

        criteria = {'exact_match': True}

        if user_id is not None:
            criteria[ 'id' ] = user_id

        if login is not None:
            criteria[ 'login' ] = self.applyTransform( login )

        view_name = createViewName('_verifyUser', user_id or login)
        keywords = createKeywords(**criteria)
        cached_info = self.ZCacheable_get( view_name=view_name
                                            , keywords=keywords
                                            , default=None
                                            )

        if cached_info is not None:
            return cached_info


        enumerators = plugins.listPlugins( IUserEnumerationPlugin )

        for enumerator_id, enumerator in enumerators:
            try:
                info = enumerator.enumerateUsers( **criteria )

                if info:
                    # Put the computed value into the cache
                    self.ZCacheable_set( info[0]
                                        , view_name=view_name
                                        , keywords=keywords
                                        )
                    return info[0]

            except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                reraise(enumerator)
                msg = 'UserEnumerationPlugin %s error' % enumerator_id
                logger.debug(msg, exc_info=True)

        return None

    security.declarePrivate( '_authorizeUser' )
    def _authorizeUser( self
                      , user
                      , accessed
                      , container
                      , name
                      , value
                      , roles=_noroles
                      ):

        """ -> boolean (whether user has roles).

        o Add the user to the SM's stack, if successful.

        o Return
        """
        user = aq_base( user ).__of__( self )
        newSecurityManager( None, user )
        security = getSecurityManager()
        try:
            try:
                if roles is _noroles:
                    if security.validate( accessed
                                        , container
                                        , name
                                        , value
                                        ):
                        return 1
                else:
                    if security.validate( accessed
                                        , container
                                        , name
                                        , value
                                        , roles
                                        ):
                        return 1
            except:
                noSecurityManager()
                raise

        except Unauthorized:
            pass

        return 0


    security.declarePrivate( '_isTop' )
    def _isTop( self ):

        """ Are we the user folder in the root object?
        """
        try:
            parent = aq_base( aq_parent( self ) )
            if parent is None:
                return 0
            return parent.isTopLevelPrincipiaApplicationObject
        except AttributeError:
            return 0


    security.declarePrivate( '_getObjectContext' )
    def _getObjectContext( self, v, request ):

        """ request -> ( a, c, n, v )

        o 'a 'is the object the object was accessed through

        o 'c 'is the physical container of the object

        o 'n 'is the name used to access the object

        o 'v' is the object (value) we're validating access to

        o XXX:  Lifted from AccessControl.User.BasicUserFolder._getobcontext
        """
        if len( request.steps ) == 0: # someone deleted root index_html

            request[ 'RESPONSE' ].notFoundError(
                'no default view (root default view was probably deleted)' )

        root = request[ 'PARENTS' ][ -1 ]
        request_container = aq_parent( root )

        n = request.steps[ -1 ]

        # default to accessed and container as v.aq_parent
        a = c = request[ 'PARENTS' ][ 0 ]

        # try to find actual container
        inner = aq_inner( v )
        innerparent = aq_parent( inner )

        if innerparent is not None:

            # this is not a method, we needn't treat it specially
            c = innerparent

        elif hasattr(v, 'im_self'):

            # this is a method, we need to treat it specially
            c = v.im_self
            c = aq_inner( v )

        # if pub's aq_parent or container is the request container, it
        # means pub was accessed from the root
        if a is request_container:
            a = root

        if c is request_container:
            c = root

        return a, c, n, v

    security.declarePrivate( '_getEmergencyUser' )
    def _getEmergencyUser( self ):

        return emergency_user.__of__( self )


    security.declarePrivate( '_doAddUser' )
    def _doAddUser( self, login, password, roles, domains, **kw ):
        """ Create a user with login, password and roles if, and only if,
            we have a registered user manager and role manager that will
            accept specific plugin interfaces.
        """
        plugins = self._getOb( 'plugins' )
        useradders = plugins.listPlugins( IUserAdderPlugin )
        roleassigners = plugins.listPlugins( IRoleAssignerPlugin )

        user = None
        login = self.applyTransform( login )

        if not (useradders and roleassigners):
            raise NotImplementedError( "There are no plugins"
                                       " that can create"
                                       " users and assign roles to them." )

        for useradder_id, useradder in useradders:
            if useradder.doAddUser( login, password ):
                # XXX: Adds user to cache, but without roles...
                user = self.getUser( login )
                break

        # No useradder was successful.
        if user is None:
            return

        for roleassigner_id, roleassigner in roleassigners:
            for role in roles:
                try:
                    roleassigner.doAssignRoleToPrincipal( user.getId(), role )
                except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                    reraise(roleassigner)
                    logger.debug( 'RoleAssigner %s error' % roleassigner_id
                                , exc_info=True
                                )
                    pass

        if user is not None:
            notify(PrincipalCreated(user))
        return user

    security.declarePublic('all_meta_types')
    def all_meta_types(self):
        """ What objects can be put in here?
        """
        import Products
        allowed_types = tuple(MultiPlugins) + (RAMCacheManager.meta_type,)

        return [x for x in Products.meta_types if x['name'] in allowed_types]

    security.declarePrivate( 'manage_beforeDelete' )
    def manage_beforeDelete(self, item, container):
        if item is self:
            try:
                del container.__allow_groups__
            except:
                pass

            handle = self.meta_type + '/' + self.getId()
            BeforeTraverse.unregisterBeforeTraverse(container, handle)

    security.declarePrivate( 'manage_afterAdd' )
    def manage_afterAdd(self, item, container):
        if item is self:
            container.__allow_groups__ = aq_base(self)

            handle = self.meta_type + '/' + self.getId()
            container = container.this()
            nc = BeforeTraverse.NameCaller(self.getId())
            BeforeTraverse.registerBeforeTraverse(container, nc, handle)

    def __call__(self, container, req):
        """ The __before_publishing_traverse__ hook.
        """
        resp = req['RESPONSE']
        req._hold(ResponseCleanup(resp))
        stack = getattr(resp, '_unauthorized_stack', [])
        stack.append(resp._unauthorized)
        resp._unauthorized_stack = stack
        resp._unauthorized = self._unauthorized
        resp._has_challenged = False

    security.declarePublic( 'applyTransform' )
    def applyTransform( self, value ):
        """ Transform for login name.

        Possibly transform the login, for example by making it lower
        case.

        value must be a string (or unicode) or it may be a sequence
        (list, tuple), in which case we need to iterate over it.
        """
        if not value:
            return value
        transform = self._get_login_transform_method()
        if not transform:
            return value
        if isinstance(value, basestring):
            return transform(value)
        result = []
        for v in value:
            result.append(transform(v))
        if isinstance(value, tuple):
            return tuple(result)
        return result

    security.declarePrivate( '_get_login_transform_method' )
    def _get_login_transform_method( self ):
        """ Get the transform method for the login name or None.
        """
        login_transform = getattr(self, 'login_transform', None)
        if not login_transform:
            return
        transform = getattr(self, login_transform.strip(), None)
        if transform is None:
            logger.debug("Transform method %r not found in plugin %r.",
                         self.login_transform, self)
            return
        return transform

    security.declarePrivate( '_setPropValue' )
    def _setPropValue(self, id, value):
        if id == 'login_transform':
            orig_value = getattr(self, id)
        super(PluggableAuthService, self)._setPropValue(id, value)
        if id == 'login_transform' and value and value != orig_value:
            logger.debug("login_transform changed from %r to %r. "
                         "Updating existing login names.", orig_value, value)
            self.updateAllLoginNames()

    security.declarePublic( 'lower' )
    def lower( self, value ):
        """ Transform for login name.

        Strip the value and lowercase it.

        To use this, set login_tranform to 'lower'.
        """
        return value.strip().lower()

    security.declarePublic( 'upper' )
    def upper( self, value ):
        """ Transform for login name.

        Strip the value and uppercase it.

        To use this, set login_tranform to 'upper'.
        """
        return value.strip().upper()

    #
    # Response override
    #
    def _unauthorized(self):
        req = self.REQUEST
        resp = req['RESPONSE']
        if resp._has_challenged: # Been here already
            return
        if not self.challenge(req, resp):
            # Need to fall back here
            resp = self._cleanupResponse()
            resp._unauthorized()
        else:
            resp._has_challenged = True

    def challenge(self, request, response):
        plugins = self._getOb('plugins')

        # Find valid protocols for this request type
        valid_protocols = []
        choosers = []
        try:
            choosers = plugins.listPlugins( IChallengeProtocolChooser )
        except KeyError:
            # Work around the fact that old instances might not have
            # IChallengeProtocolChooser registered with the
            # PluginRegistry.
            pass

        for chooser_id, chooser in choosers:
            choosen = chooser.chooseProtocols(request)
            if choosen is None:
                continue
            valid_protocols.extend(choosen)

        # Go through all challenge plugins
        challengers = plugins.listPlugins( IChallengePlugin )

        protocol = None

        for challenger_id, challenger in challengers:
            challenger_protocol = getattr(challenger, 'protocol',
                                          challenger_id)
            if valid_protocols and challenger_protocol not in valid_protocols:
                # Skip invalid protocol for this request type.
                continue
            if protocol is None or protocol == challenger_protocol:
                if challenger.challenge(request, response):
                    protocol = challenger_protocol

        if protocol is not None:
            # something fired, so it was a successful PAS challenge
            return True

        # nothing fired, so trigger the fallback
        return False

    def _cleanupResponse(self):
        resp = self.REQUEST['RESPONSE']
        # No errors of any sort may propagate, and we don't care *what*
        # they are, even to log them.
        stack = getattr(resp, '_unauthorized_stack', [])

        if stack:
            resp._unauthorized = stack.pop()
        else:
            try:
                del resp._unauthorized
            except:
                pass

        return resp

    security.declarePublic( 'hasUsers' )
    def hasUsers(self):
        """Zope quick start sacrifice.

        The quick start page expects a hasUsers() method.
        """
        return True

    security.declarePublic('updateCredentials')
    def updateCredentials(self, request, response, login, new_password):
        """Central updateCredentials method

        This method is needed for cases where the credentials storage and
        the credentials extraction is handled by different plugins. Example
        case would be if the CookieAuthHelper is used as a Challenge and
        Extraction plugin only to take advantage of the login page feature
        but the credentials are not stored in the CookieAuthHelper cookie
        but somewhere else, like in a Session.
        """
        login = self.applyTransform(login)
        plugins = self._getOb('plugins')
        cred_updaters = plugins.listPlugins(ICredentialsUpdatePlugin)

        for updater_id, updater in cred_updaters:
            updater.updateCredentials(request, response, login, new_password)

    security.declarePublic('logout')
    def logout(self, REQUEST):
        """Publicly accessible method to log out a user
        """
        self.resetCredentials(REQUEST, REQUEST['RESPONSE'])

        # Little bit of a hack: Issuing a redirect to the same place
        # where the user was so that in the second request the now-destroyed
        # credentials can be acted upon to e.g. go back to the login page
        referrer = REQUEST.get('HTTP_REFERER') # optional header
        if referrer:
            REQUEST['RESPONSE'].redirect(referrer)

    security.declarePublic('resetCredentials')
    def resetCredentials(self, request, response):
        """Reset credentials by informing all active resetCredentials plugins
        """
        user = getSecurityManager().getUser()
        if aq_base(user) is not nobody:
            plugins = self._getOb('plugins')
            cred_resetters = plugins.listPlugins(ICredentialsResetPlugin)

            for resetter_id, resetter in cred_resetters:
                resetter.resetCredentials(request, response)


    security.declareProtected( ManageUsers, 'updateLoginName')
    def updateLoginName(self, user_id, login_name):
        """Update login name of user.
        """
        logger.debug("Called updateLoginName, user_id=%r, login_name=%r",
                     user_id, login_name)
        login_name = self.applyTransform(login_name)
        # Note: we do not call self.getUserById(user_id) here to see
        # if this user exists, or call getUserName() on the answer to
        # compare it with the transformed login name, because the user
        # may be reported with a login name that is transformed on the
        # fly, for example in the _verifyUser call, even though the
        # plugin has not actually transformed the login name yet in
        # the backend.
        self._updateLoginName(user_id, login_name)

    security.declarePublic('updateOwnLoginName')
    def updateOwnLoginName(self, login_name):
        """Update own login name of authenticated user.
        """
        logger.debug("Called updateOwnLoginName, login_name=%r", login_name)
        login_name = self.applyTransform(login_name)
        user = getSecurityManager().getUser()
        if aq_base(user) is nobody:
            return
        user_id = user.getId()
        # Note: we do not compare the login name here.  See the
        # comment in updateLoginName above.
        self._updateLoginName(user_id, login_name)

    def _updateLoginName(self, user_id, login_name):
        # Note: we do not compare the login name here.  See the
        # comment in updateLoginName above.
        plugins = self._getOb('plugins')
        updaters = plugins.listPlugins(IUserEnumerationPlugin)

        # Call the updaters.  One of them *must* succeed without an
        # exception, even if it does not change anything.  When a
        # login name is already taken, we do not want to fail
        # silently.
        success = False
        for updater_id, updater in updaters:
            if not hasattr(updater, 'updateUser'):
                # This was a later addition to the interface, so we
                # are forgiving.
                logger.warn("%s plugin lacks updateUser method of "
                            "IUserEnumerationPlugin.", updater_id)
                continue
            try:
                result = updater.updateUser(user_id, login_name)
            except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                reraise(updater)
                logger.debug('UpdateLoginNamePlugin %s error' % updater_id,
                             exc_info=True)
            else:
                if result:
                    success = True
                    logger.debug("login name changed to: %r", login_name)

        if not success:
            raise ValueError("Cannot update login name of user %r to %r. "
                             "Possibly duplicate." % (user_id, login_name))

    security.declareProtected( ManageUsers, 'updateLoginName')
    def updateAllLoginNames(self, quit_on_first_error=True):
        """Update login names of all users to their canonical value.

        This should be done after changing the login_transform
        property of PAS.

        You can set quit_on_first_error to False to report all errors
        before quitting with an error.  This can be useful if you want
        to know how many problems there are, if any.
        """
        plugins = self._getOb('plugins')
        updaters = plugins.listPlugins(IUserEnumerationPlugin)
        for updater_id, updater in updaters:
            if not hasattr(updater, 'updateEveryLoginName'):
                # This was a later addition to the interface, so we
                # are forgiving.
                logger.warn("%s plugin lacks updateEveryLoginName method of "
                            "IUserEnumerationPlugin.", updater_id)
                continue
            # Note: do not swallow any exceptions here.
            updater.updateEveryLoginName(
                quit_on_first_error=quit_on_first_error)


classImplements( PluggableAuthService
               , (IPluggableAuthService, IObjectManager, IPropertyManager)
               )

InitializeClass( PluggableAuthService )


class ResponseCleanup:
    def __init__(self, resp):
        self.resp = resp

    def __del__(self):
        # Free the references.
        #
        # No errors of any sort may propagate, and we don't care *what*
        # they are, even to log them.
        stack = getattr(self.resp, '_unauthorized_stack', [])
        old = None

        while stack:
            old = stack.pop()

        if old is not None:
            self.resp._unauthorized = old
        else:
            try:
                del self.resp._unauthorized
            except:
                pass

        try:
            del self.resp
        except:
            pass

_PLUGIN_TYPE_INFO = (
    ( IExtractionPlugin
    , 'IExtractionPlugin'
    , 'extraction'
    , "Extraction plugins are responsible for extracting credentials "
      "from the request."
    )
  , ( IAuthenticationPlugin
    , 'IAuthenticationPlugin'
    , 'authentication'
    , "Authentication plugins are responsible for validating credentials "
      "generated by the Extraction Plugin."
    )
  , ( IChallengePlugin
    , 'IChallengePlugin'
    , 'challenge'
    , "Challenge plugins initiate a challenge to the user to provide "
      "credentials."
    )
  , ( ICredentialsUpdatePlugin
    , 'ICredentialsUpdatePlugin'
    , 'update credentials'
    , "Credential update plugins respond to the user changing "
      "credentials."
    )
  , ( ICredentialsResetPlugin
    , 'ICredentialsResetPlugin'
    , 'reset credentials'
    , "Credential clear plugins respond to a user logging out."
    )
  , ( IUserFactoryPlugin
    , 'IUserFactoryPlugin'
    , 'userfactory'
    , "Create users."
    )
  , ( IAnonymousUserFactoryPlugin
    , 'IAnonymousUserFactoryPlugin'
    , 'anonymoususerfactory'
    , "Create anonymous users."
    )
  , ( IPropertiesPlugin
    , 'IPropertiesPlugin'
    , 'properties'
    , "Properties plugins generate property sheets for users."
    )
  , ( IGroupsPlugin
    , 'IGroupsPlugin'
    , 'groups'
    , "Groups plugins determine the groups to which a user belongs."
    )
  , ( IRolesPlugin
    , 'IRolesPlugin'
    , 'roles'
    , "Roles plugins determine the global roles which a user has."
    )
  , ( IUpdatePlugin
    , 'IUpdatePlugin'
    , 'update'
    , "Update plugins allow the user or the application to update "
      "the user's properties."
    )
  , ( IValidationPlugin
    , 'IValidationPlugin'
    , 'validation'
    , "Validation plugins specify allowable values for user properties "
      "(e.g., minimum password length, allowed characters, etc.)"
    )
  , ( IUserEnumerationPlugin
    , 'IUserEnumerationPlugin'
    , 'user_enumeration'
    , "Enumeration plugins allow querying users by ID, and searching for "
      "users who match particular criteria."
    )
  , ( IUserAdderPlugin
    , 'IUserAdderPlugin'
    , 'user_adder'
    , "User Adder plugins allow the Pluggable Auth Service to create users."
    )
  , ( IGroupEnumerationPlugin
    , 'IGroupEnumerationPlugin'
    , 'group_enumeration'
    , "Enumeration plugins allow querying groups by ID."
    )
  , ( IRoleEnumerationPlugin
    , 'IRoleEnumerationPlugin'
    , 'role_enumeration'
    , "Enumeration plugins allow querying roles by ID."
    )
  , ( IRoleAssignerPlugin
    , 'IRoleAssignerPlugin'
    , 'role_assigner'
    , "Role Assigner plugins allow the Pluggable Auth Service to assign"
      " roles to principals."
    )
  , ( IChallengeProtocolChooser
    , 'IChallengeProtocolChooser'
    , 'challenge_protocol_chooser'
    , "Challenge Protocol Chooser plugins decide what authorization"
      "protocol to use for a given request type."
    )
  , ( IRequestTypeSniffer
    , 'IRequestTypeSniffer'
    , 'request_type_sniffer'
    , "Request Type Sniffer plugins detect the type of an incoming request."
    )
  , ( INotCompetentPlugin
    , 'INotCompetentPlugin'
    , 'notcompetent'
    , "Not-Competent plugins check whether this user folder should not "
      "authenticate the current request. "
      "These plugins are not used for a top level user folder. "
      "They are typically used to prevent shaddowing of authentications by "
      "higher level user folders."
    )
  )

def addPluggableAuthService( dispatcher
                           , base_profile=None
                           , extension_profiles=()
                           , create_snapshot=True
                           , setup_tool_id='setup_tool'
                           , REQUEST=None
                           ):
    """ Add a PluggableAuthService to 'dispatcher'.

    o BBB for non-GenericSetup use.
    """
    pas = PluggableAuthService()
    preg = PluginRegistry( _PLUGIN_TYPE_INFO )
    preg._setId( 'plugins' )
    pas._setObject( 'plugins', preg )
    dispatcher._setObject( pas.getId(), pas )

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(
                                '%s/manage_workspace'
                                '?manage_tabs_message='
                                'PluggableAuthService+added.'
                              % dispatcher.absolute_url() )

def addConfiguredPASForm(dispatcher):
    """ Wrap the PTF in 'dispatcher', including 'profile_registry' in options.
    """
    from Products.GenericSetup import EXTENSION
    from Products.GenericSetup import profile_registry

    wrapped = PageTemplateFile( 'pasAddForm', _wwwdir ).__of__( dispatcher )

    base_profiles = []
    extension_profiles = []

    for info in profile_registry.listProfileInfo(for_=IPluggableAuthService):
        if info.get('type') == EXTENSION:
            extension_profiles.append(info)
        else:
            base_profiles.append(info)

    return wrapped( base_profiles=tuple(base_profiles),
                    extension_profiles =tuple(extension_profiles) )

def addConfiguredPAS( dispatcher
                    , base_profile
                    , extension_profiles=()
                    , create_snapshot=True
                    , setup_tool_id='setup_tool'
                    , REQUEST=None
                    ):
    """ Add a PluggableAuthService to 'self.
    """
    from Products.GenericSetup.tool import SetupTool

    pas = PluggableAuthService()
    preg = PluginRegistry( _PLUGIN_TYPE_INFO )
    preg._setId( 'plugins' )
    pas._setObject( 'plugins', preg )
    dispatcher._setObject( pas.getId(), pas )

    pas = dispatcher._getOb( pas.getId() )    # wrapped
    tool = SetupTool( setup_tool_id )
    pas._setObject( tool.getId(), tool )

    tool = pas._getOb( tool.getId() )       # wrapped
    tool.setImportContext( 'profile-%s' % base_profile )
    tool.runAllImportSteps()

    for extension_profile in extension_profiles:
        tool.setImportContext( 'profile-%s' % extension_profile )
        tool.runAllImportSteps()

    tool.setImportContext( 'profile-%s' % base_profile )

    if create_snapshot:
        tool.createSnapshot( 'initial_configuration' )

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(
                                '%s/manage_workspace'
                                '?manage_tabs_message='
                                'PluggableAuthService+added.'
                              % dispatcher.absolute_url() )




##############################################################################
#
# Copyright (c) 2006 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import os, sys, base64, unittest

from Products.PluggableAuthService.tests import pastc

from AccessControl import Unauthorized
from AccessControl.Permissions import view as View
from AccessControl.Permissions import add_folders as AddFolders

from Products.PluggableAuthService.PluggableAuthService import PluggableAuthService

from zope import event
from zope.component import adapter
from zope.component import provideHandler
from Products.PluggableAuthService.interfaces.events import IPrincipalCreatedEvent
from Products.PluggableAuthService.events import CredentialsUpdated
from Products.PluggableAuthService.events import PASEventNotify
from Products.PluggableAuthService.events import userCredentialsUpdatedHandler

@adapter(IPrincipalCreatedEvent)
def userCreatedHandler(event):
    pas = event.principal.aq_parent
    if not hasattr(pas, 'events'):
        pas.events = []

    pas.events.append(event)


class UserFolderTests(pastc.PASTestCase):

    def afterSetUp(self):
        # Set up roles and a user
        self.uf = self.folder.acl_users
        self.folder._addRole('role1')
        self.folder.manage_role('role1', [View])
        self.uf.roles.addRole('role1')
        self.folder._addRole('role2')
        self.folder.manage_role('role2', [View])
        self.uf.roles.addRole('role2')
        self.uf._doAddUser('user1', 'secret', ['role1'], [])
        # Set up a published object accessible to user
        self.folder.addDTMLMethod('doc', file='the document')
        self.doc = self.folder.doc
        self.doc.manage_permission(View, ['role1'], acquire=0)
        # Rig the REQUEST so it looks like we traversed to doc
        self.app.REQUEST['PUBLISHED'] = self.doc
        self.app.REQUEST['PARENTS'] = [self.app, self.folder]
        self.app.REQUEST.steps = list(self.doc.getPhysicalPath())
        self.basic = 'Basic %s' % base64.encodestring('user1:secret').rstrip()
        # Make sure we are not logged in
        self.logout()

    def testGetUser(self):
        self.failIfEqual(self.uf.getUser('user1'), None)

    def testGetBadUser(self):
        self.assertEqual(self.uf.getUser('user2'), None)

    def testGetUserById(self):
        self.failIfEqual(self.uf.getUserById('user1'), None)

    def testGetBadUserById(self):
        self.assertEqual(self.uf.getUserById('user2'), None)

    def NOTIMPLEMENTED_testGetUsers(self):
        users = self.uf.getUsers()
        self.failUnless(users)
        self.assertEqual(users[0].getUserName(), 'user1')

    def NOTIMPLEMENTED_testGetUserNames(self):
        names = self.uf.getUserNames()
        self.failUnless(names)
        self.assertEqual(names[0], 'user1')

    def NOTIMPLEMENTED_testIdentify(self):
        name, password = self.uf.identify(self.basic)
        self.assertEqual(name, 'user1')
        self.assertEqual(password, 'secret')

    def testGetRoles(self):
        user = self.uf.getUser('user1')
        self.failUnless('role1' in user.getRoles())
        self.failIf('role2' in user.getRoles())

    def testGetRolesInContext(self):
        user = self.uf.getUser('user1')
        self.folder.manage_addLocalRoles('user1', ['role2'])
        roles = user.getRolesInContext(self.folder)
        self.failUnless('role1' in roles)
        self.failUnless('role2' in roles)

    def testHasRole(self):
        user = self.uf.getUser('user1')
        self.failUnless(user.has_role('role1', self.folder))

    def testHasLocalRole(self):
        user = self.uf.getUser('user1')
        self.failIf(user.has_role('role2', self.folder))
        self.folder.manage_addLocalRoles('user1', ['role2'])
        self.failUnless(user.has_role('role2', self.folder))

    def testHasPermission(self):
        user = self.uf.getUser('user1')
        self.failUnless(user.has_permission(View, self.folder))
        self.failIf(user.has_permission(AddFolders, self.folder))
        self.folder.manage_role('role1', [AddFolders])
        self.failUnless(user.has_permission(AddFolders, self.folder))

    def testHasLocalRolePermission(self):
        user = self.uf.getUser('user1')
        self.folder.manage_role('role2', [AddFolders])
        self.failIf(user.has_permission(AddFolders, self.folder))
        self.folder.manage_addLocalRoles('user1', ['role2'])
        self.failUnless(user.has_permission(AddFolders, self.folder))

    def NOTIMPLEMENTED_testAuthenticate(self):
        user = self.uf.getUser('user1')
        self.failUnless(user.authenticate('secret', self.app.REQUEST))

    def testValidate(self):
        # XXX: PAS validate ignores auth argument
        self.app.REQUEST._auth = self.basic
        user = self.uf.validate(self.app.REQUEST, self.basic, ['role1'])
        self.failIfEqual(user, None)
        self.assertEqual(user.getUserName(), 'user1')

    def testNotValidateWithoutAuth(self):
        # XXX: PAS validate ignores auth argument
        user = self.uf.validate(self.app.REQUEST, '', ['role1'])
        self.assertEqual(user, None)

    def testValidateWithoutRoles(self):
        # Note - calling uf.validate without specifying roles will cause
        # the security machinery to determine the needed roles by looking
        # at the object itself (or its container). I'm putting this note
        # in to clarify because the original test expected failure but it
        # really should have expected success, since the user and the
        # object being checked both have the role 'role1', even though no
        # roles are passed explicitly to the userfolder validate method.
        # XXX: PAS validate ignores auth argument
        self.app.REQUEST._auth = self.basic
        user = self.uf.validate(self.app.REQUEST, self.basic)
        self.assertEqual(user.getUserName(), 'user1')

    def testNotValidateWithEmptyRoles(self):
        # XXX: PAS validate ignores auth argument
        self.app.REQUEST._auth = self.basic
        user = self.uf.validate(self.app.REQUEST, self.basic, [])
        self.assertEqual(user, None)

    def testNotValidateWithWrongRoles(self):
        # XXX: PAS validate ignores auth argument
        self.app.REQUEST._auth = self.basic
        user = self.uf.validate(self.app.REQUEST, self.basic, ['role2'])
        self.assertEqual(user, None)

    def testAllowAccessToUser(self):
        self.login('user1')
        try:
            self.folder.restrictedTraverse('doc')
        except Unauthorized:
            self.fail('Unauthorized')

    def testDenyAccessToAnonymous(self):
        self.assertRaises(Unauthorized, self.folder.restrictedTraverse, 'doc')

    def testMaxListUsers(self):
        # create a folder-ish thing which contains a roleManager,
        # then put an acl_users object into the folde-ish thing

        class Folderish(PluggableAuthService):
            def __init__(self, size, count):
                self.maxlistusers = size
                self.users = []
                self.acl_users = self
                self.__allow_groups__ = self
                for i in xrange(count):
                    self.users.append("Nobody")

            def getUsers(self):
                return self.users

            def user_names(self):
                return self.getUsers()


        tinyFolderOver = Folderish(15, 20)
        tinyFolderUnder = Folderish(15, 10)

        assert tinyFolderOver.maxlistusers == 15
        assert tinyFolderUnder.maxlistusers == 15
        assert len(tinyFolderOver.user_names()) == 20
        assert len(tinyFolderUnder.user_names()) == 10

        try:
            list = tinyFolderOver.get_valid_userids()
            assert 0, "Did not raise overflow error"
        except OverflowError:
            pass

        try:
            list = tinyFolderUnder.get_valid_userids()
            pass
        except OverflowError:
            assert 0, "Raised overflow error erroneously"

    def test__doAddUser_with_not_yet_encrypted_passwords(self):
        # See collector #1869 && #1926
        from AccessControl.AuthEncoding import is_encrypted

        USER_ID = 'not_yet_encrypted'
        PASSWORD = 'password'

        self.failIf(is_encrypted(PASSWORD))

        self.uf._doAddUser(USER_ID, PASSWORD, [], [])

        uid_and_info = self.uf.users.authenticateCredentials(
                                { 'login': USER_ID
                                , 'password': PASSWORD
                                })

        self.assertEqual(uid_and_info, (USER_ID, USER_ID))

    def test__doAddUser_with_preencrypted_passwords(self):
        # See collector #1869 && #1926
        from AccessControl.AuthEncoding import pw_encrypt

        USER_ID = 'already_encrypted'
        PASSWORD = 'password'

        ENCRYPTED = pw_encrypt(PASSWORD)

        self.uf._doAddUser(USER_ID, ENCRYPTED, [], [])

        uid_and_info = self.uf.users.authenticateCredentials(
                                { 'login': USER_ID
                                , 'password': PASSWORD
                                })

        self.assertEqual(uid_and_info, (USER_ID, USER_ID))

    def test_manage_zmi_logout(self):
        request = self.app.REQUEST
        response = request.RESPONSE
        self.folder.manage_zmi_logout(request, response)
        self.assertEqual(response.status, 401)
        self.assertEqual(response.headers.get('WWW-Authenticate'),
                         'basic realm="%s"' % response.realm)


class UserTests(pastc.PASTestCase):

    def afterSetUp(self):
        self.uf = self.folder.acl_users
        self.uf._doAddUser('chris', '123', ['Manager'], [])
        self.user = self.uf.getUser('chris')

    def testGetUserName(self):
        f = self.user
        self.assertEqual(f.getUserName(), 'chris')

    def testGetUserId(self):
        f = self.user
        self.assertEqual(f.getId(), 'chris')

    def testBaseUserGetIdEqualGetName(self):
        # this is true for the default user type, but will not
        # always be true for extended user types going forward (post-2.6)
        f = self.user
        self.assertEqual(f.getId(), f.getUserName())

    def NOTIMPLEMENTED_testGetPassword(self):
        f = self.user
        self.assertEqual(f._getPassword(), '123')

    def testGetRoles(self):
        f = self.user
        # XXX: PAS returns roles as list
        #self.assertEqual(f.getRoles(), ('Manager', 'Authenticated'))
        self.assertEqual(f.getRoles(), ['Manager', 'Authenticated'])

    def testGetDomains(self):
        f = self.user
        self.assertEqual(f.getDomains(), ())


class UserEvents(pastc.PASTestCase):

    def afterSetUp(self):
        # Set up roles and a user
        self.uf = self.folder.acl_users
        self.folder._addRole('role1')
        self.folder.manage_role('role1', [View])
        self.uf.roles.addRole('role1')
        self.folder._addRole('role2')
        self.uf._doAddUser('user1', 'secret', ['role1'], [])

    def testUserCreationEvent(self):
        provideHandler(userCreatedHandler)
        self.uf.events = []

        self.uf._doAddUser('event1', 'secret', ['role1'], [])

        self.assertEqual(len(self.uf.events), 1)
        event = self.uf.events[0]
        self.failUnless(IPrincipalCreatedEvent.providedBy(event))
        self.assertEqual(event.principal.getUserName(), 'event1')
        self.assertEqual(event.principal.getId(), 'event1')

    def testCredentialsEvent(self):
        provideHandler(PASEventNotify)
        provideHandler(userCredentialsUpdatedHandler)
        def wrap(self, *args):
            self._data.append(args)
            return self._original(*args)
        self.uf._data=[]
        self.uf._original=self.uf.updateCredentials
        self.uf.updateCredentials=wrap
        self.assertEqual(len(self.uf._data), 0)
        event.notify(CredentialsUpdated(self.uf.getUserById("user1"), "testpassword"))
        self.assertEqual(self.uf._data[0][2], "user1")
        self.assertEqual(self.uf._data[0][3], "testpassword")



def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(UserFolderTests))
    suite.addTest(unittest.makeSuite(UserTests))
    suite.addTest(unittest.makeSuite(UserEvents))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest
import os.path
import sys

from DateTime.DateTime import DateTime

from OFS.Image import Image

from conformance import IPropertySheet_conformance

path = os.path.dirname(__file__)
path = os.path.join(path, 'image.gif')
img_file = open(path, 'r')

class UserPropertySheetTests( unittest.TestCase
                            , IPropertySheet_conformance
                            ):

    _SCHEMA = ( ( 's', 'string'  )
              , ( 'i', 'int'     )
              , ( 'f', 'float'   )
              , ( 'n', 'long'    )
              , ( 'd', 'date'    )
              , ( 'l', 'lines'   )
              , ( 't', 'lines'   )
              , ( 'b', 'boolean' )
              , ( 'img', 'image' )
              )

    _STRING_VALUE = 'string'
    _INT_VALUE = 42
    _FLOAT_VALUE = 9.8
    _LONG_VALUE = sys.maxint + 1
    _DATE_VALUE = DateTime()
    _LIST_VALUE = [ 'a', 'b', 'c' ]
    _TUPLE_VALUE = ( 'd', 'e', 'f' )
    _BOOL_VALUE = True
    _IMG_VALUE = Image('image', 'Test Image', img_file)

    def _getTargetClass( self ):

        from Products.PluggableAuthService.UserPropertySheet \
            import UserPropertySheet

        return UserPropertySheet

    def _makeOne( self, *args, **kw ):

        return self._getTargetClass()( *args, **kw )

    def test_ctor_id_noschema_novalues( self ):

        ups = self._makeOne( 'empty' )

        self.assertEqual( ups.getId(), 'empty' )

        self.failIf( ups.hasProperty( 'empty' ) )
        self.failIf( ups.hasProperty( 'foo' ) )
        self.failIf( ups.hasProperty( 'bar' ) )

        self.assertEqual( ups.getProperty( 'foo' ), None )
        self.assertEqual( ups.getPropertyType( 'foo' ), None )

        self.assertEqual( len( ups.propertyMap() ), 0 )
        self.assertEqual( len( ups.propertyIds() ), 0 )
        self.assertEqual( len( ups.propertyValues() ), 0 )
        self.assertEqual( len( ups.propertyItems() ), 0 )
        self.assertEqual( len( ups.propertyIds() ), 0 )

    def _checkStockSchema( self, ups, values_are_none=False ):

        self.failIf( ups.hasProperty( 'x' ) )
        self.failUnless( ups.hasProperty( 's' ) )
        self.failUnless( ups.hasProperty( 'i' ) )
        self.failUnless( ups.hasProperty( 'f' ) )
        self.failUnless( ups.hasProperty( 'n' ) )
        self.failUnless( ups.hasProperty( 'd' ) )
        self.failUnless( ups.hasProperty( 'l' ) )
        self.failUnless( ups.hasProperty( 't' ) )
        self.failUnless( ups.hasProperty( 'b' ) )
        self.failUnless( ups.hasProperty( 'img' ) )

        self.assertEqual( ups.getPropertyType( 's' ), 'string' )
        self.assertEqual( ups.propertyInfo( 's' )[ 'type' ], 'string' )
        if values_are_none:
            self.assertEqual( ups.getProperty( 's' ), None )
        else:
            self.assertEqual( ups.getProperty( 's' ), self._STRING_VALUE )

        self.assertEqual( ups.getPropertyType( 'i' ), 'int' )
        self.assertEqual( ups.propertyInfo( 'i' )[ 'type' ], 'int' )
        if values_are_none:
            self.assertEqual( ups.getProperty( 'i' ), None )
        else:
            self.assertEqual( ups.getProperty( 'i' ), self._INT_VALUE )

        self.assertEqual( ups.getPropertyType( 'f' ), 'float' )
        self.assertEqual( ups.propertyInfo( 'f' )[ 'type' ], 'float' )
        if values_are_none:
            self.assertEqual( ups.getProperty( 'f' ), None )
        else:
            self.assertEqual( ups.getProperty( 'f' ), self._FLOAT_VALUE )

        self.assertEqual( ups.getPropertyType( 'n' ), 'long' )
        self.assertEqual( ups.propertyInfo( 'n' )[ 'type' ], 'long' )
        if values_are_none:
            self.assertEqual( ups.getProperty( 'n' ), None )
        else:
            self.assertEqual( ups.getProperty( 'n' ), self._LONG_VALUE )

        self.assertEqual( ups.getPropertyType( 'd' ), 'date' )
        self.assertEqual( ups.propertyInfo( 'd' )[ 'type' ], 'date' )
        if values_are_none:
            self.assertEqual( ups.getProperty( 'd' ), None )
        else:
            self.assertEqual( ups.getProperty( 'd' ), self._DATE_VALUE )

        self.assertEqual( ups.getPropertyType( 'b' ), 'boolean' )
        self.assertEqual( ups.propertyInfo( 'b' )[ 'type' ], 'boolean' )
        if values_are_none:
            self.assertEqual( ups.getProperty( 'b' ), None )
        else:
            self.assertEqual( ups.getProperty( 'b' ), self._BOOL_VALUE )

        self.assertEqual( ups.getPropertyType( 'l' ), 'lines' )
        self.assertEqual( ups.propertyInfo( 'l' )[ 'type' ], 'lines' )

        if values_are_none:
            self.assertEqual( ups.getProperty( 'l' ), None )
        else:
            got = ups.getProperty( 'l' )
            self.assertEqual( type( got ), type( () ) )
            self.assertEqual( len( got ), len( self._LIST_VALUE ) )

            for i in range( len( self._LIST_VALUE ) ):
                self.assertEqual( got[i], self._LIST_VALUE[i] )

        self.assertEqual( ups.getPropertyType( 't' ), 'lines' )
        self.assertEqual( ups.propertyInfo( 't' )[ 'type' ], 'lines' )

        if values_are_none:
            self.assertEqual( ups.getProperty( 't' ), None )
        else:
            got = ups.getProperty( 't' )
            self.assertEqual( type( got ), type( () ) )
            self.assertEqual( len( got ), len( self._TUPLE_VALUE ) )

            for i in range( len( self._TUPLE_VALUE ) ):
                self.assertEqual( got[i], self._TUPLE_VALUE[i] )

        self.assertEqual( ups.getPropertyType( 'img' ), 'image' )
        self.assertEqual( ups.propertyInfo( 'img' )[ 'type' ], 'image' )

        if values_are_none:
            self.assertEqual( ups.getProperty( 'img' ), None )
        else:
            got = ups.getProperty( 'img' )
            self.assertEqual( type( got ), Image )
            self.assertEqual( got.size, self._IMG_VALUE.size )
            self.assertEqual( got, self._IMG_VALUE )


        pmap = ups.propertyMap()
        self.assertEqual( len( pmap ), len( self._SCHEMA ) )

        for i in range( len( pmap ) ):
            info = pmap[i]
            spec = [ x for x in self._SCHEMA if x[0] == info['id' ] ][0]
            self.assertEqual( info[ 'id' ], spec[0] )
            self.assertEqual( info[ 'type' ], spec[1] )
            self.assertEqual( info[ 'mode' ], '' )  # readonly, no delete


    def test_ctor__guessSchema( self ):

        ups = self._makeOne( 'guessed'
                           , s=self._STRING_VALUE
                           , i=self._INT_VALUE
                           , f=self._FLOAT_VALUE
                           , n=self._LONG_VALUE
                           , d=self._DATE_VALUE
                           , l=self._LIST_VALUE
                           , t=self._TUPLE_VALUE
                           , b=self._BOOL_VALUE
                           , img=self._IMG_VALUE
                           )

        self._checkStockSchema( ups )


    def test_ctor_w_schema(self):

        ups = self._makeOne( 'w_schema'
                           , self._SCHEMA
                           , s=self._STRING_VALUE
                           , i=self._INT_VALUE
                           , f=self._FLOAT_VALUE
                           , n=self._LONG_VALUE
                           , d=self._DATE_VALUE
                           , l=self._LIST_VALUE
                           , t=self._TUPLE_VALUE
                           , b=self._BOOL_VALUE
                           , img=self._IMG_VALUE
                           )

        self._checkStockSchema( ups )


    def test_ctor_w_schema_no_values(self):

        ups = self._makeOne( 'w_schema'
                           , self._SCHEMA
                           )

        self._checkStockSchema( ups, values_are_none=True )

if __name__ == "__main__":
    unittest.main()

def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite( UserPropertySheetTests ),
        ))



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for plugin.exportimport

$Id$
"""

import unittest

try:
    import Products.GenericSetup
except ImportError:  # No GenericSetup, so no tests

    print 'XXXX:  No GenericSetup!'
    def test_suite():
        return unittest.TestSuite()

else:

    from Products.GenericSetup.tests.conformance \
            import ConformsToIFilesystemExporter
    from Products.GenericSetup.tests.conformance \
            import ConformsToIFilesystemImporter

    from Products.GenericSetup.tests.common import BaseRegistryTests
    from Products.GenericSetup.tests.common import DummyExportContext
    from Products.GenericSetup.tests.common import DummyImportContext

    from Products.PluggableAuthService.tests.utils import _setUpDefaultTraversable

    class _TestBase(BaseRegistryTests,
                    ConformsToIFilesystemExporter,
                    ConformsToIFilesystemImporter,
                    ):

        def _makeOne(self, context, *args, **kw):
            return self._getTargetClass()(context, *args, **kw)

    class ZODBUserManagerExportImportTests(_TestBase):

        def _getTargetClass(self):
            from Products.PluggableAuthService.plugins.exportimport \
                import ZODBUserManagerExportImport
            return ZODBUserManagerExportImport

        def _makePlugin(self, id='zodbusers', *args, **kw):
            from Products.PluggableAuthService.plugins.ZODBUserManager \
                import ZODBUserManager
            return ZODBUserManager(id, *args, **kw)

        def test_listExportableItems(self):
            plugin = self._makePlugin('lEI').__of__(self.root)
            adapter = self._makeOne(plugin)
            self.assertEqual(len(adapter.listExportableItems()), 0)
            plugin.addUser('foo', 'bar', 'baz')
            self.assertEqual(len(adapter.listExportableItems()), 0)

        def test__getExportInfo_empty(self):
            plugin = self._makePlugin('empty').__of__(self.root)
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], None)
            self.assertEqual(len(info['users']), 0)

        def test_export_empty(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('empty').__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len( context._wrote ), 1 )
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual( filename, 'plugins/empty.xml' )
            self._compareDOM( text, _EMPTY_ZODB_USERS )
            self.assertEqual( content_type, 'text/xml' )

        def test__getExportInfo_with_users(self):

            plugin = self._makePlugin('with_users').__of__(self.root)
            plugin.title = 'Plugin Title'
            source_info = []

            for info in _ZODB_USER_INFO:
                info = info.copy()
                plugin.addUser(**info)
                hash = plugin._user_passwords[info['user_id']]
                info['password_hash'] = hash
                source_info.append(info)

            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], 'Plugin Title')
            self.assertEqual(len(info['users']), len(source_info))

            for x, y in zip(info['users'], source_info):
                self.assertEqual(x['user_id'], y['user_id'])
                self.assertEqual(x['login_name'], y['login_name'])
                self.assertEqual(x['password_hash'], y['password_hash'])

        def test_export_with_users(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('with_users').__of__(self.root)
            plugin.title = 'Plugin Title'

            hashes = []
            for info in _ZODB_USER_INFO:
                plugin.addUser(**info)
                hash = plugin._user_passwords[info['user_id']]
                hashes.append(hash)

            adapter = self._makeOne(plugin)
            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len(context._wrote), 1)
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual(filename, 'plugins/with_users.xml')
            self._compareDOM(text, _FILLED_ZODB_USERS % tuple(hashes))
            self.assertEqual(content_type, 'text/xml')

        def test_import_empty(self):
            HASHES = ('abcde', 'wxyz')
            plugin = self._makePlugin('empty').__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin)
            context._files['plugins/empty.xml'] = _FILLED_ZODB_USERS % HASHES
            self.assertEqual(plugin.title, None)

            adapter.import_(context, 'plugins', False)

            self.assertEqual(len(plugin._user_passwords), len(_ZODB_USER_INFO))
            self.assertEqual(plugin.title, 'Plugin Title')

            for info, hash in zip(_ZODB_USER_INFO, HASHES):
                user_id = info['user_id']
                self.assertEqual(plugin.getLoginForUserId(user_id),
                                info['login_name'])
                self.assertEqual(plugin._user_passwords[user_id], hash)

        def test_import_without_purge_leaves_existing_users(self):

            plugin = self._makePlugin('with_users').__of__(self.root)
            plugin.title = 'Plugin Title'

            for info in _ZODB_USER_INFO:
                plugin.addUser(**info)

            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, purge=False)
            context._files['plugins/with_users.xml'] = _EMPTY_ZODB_USERS

            self.assertEqual(len(plugin._user_passwords), len(_ZODB_USER_INFO))
            adapter.import_(context, 'plugins', False)
            self.assertEqual(len(plugin._user_passwords), len(_ZODB_USER_INFO))
            self.assertEqual(plugin.title, None)

        def test_import_with_purge_wipes_existing_users(self):

            plugin = self._makePlugin('with_users').__of__(self.root)
            plugin.title = 'Plugin Title'

            for info in _ZODB_USER_INFO:
                plugin.addUser(**info)

            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, purge=True)
            context._files['plugins/with_users.xml'] = _EMPTY_ZODB_USERS

            self.assertEqual(len(plugin._user_passwords), len(_ZODB_USER_INFO))
            adapter.import_(context, 'plugins', False)
            self.assertEqual(len(plugin._user_passwords), 0)
            self.assertEqual(plugin.title, None)

    class ZODBGroupManagerExportImportTests(_TestBase):

        def _getTargetClass(self):
            from Products.PluggableAuthService.plugins.exportimport \
                import ZODBGroupManagerExportImport
            return ZODBGroupManagerExportImport

        def _makePlugin(self, id, *args, **kw):
            from Products.PluggableAuthService.plugins.ZODBGroupManager \
                import ZODBGroupManager
            return ZODBGroupManager(id, *args, **kw)

        def test_listExportableItems(self):
            plugin = self._makePlugin('lEI').__of__(self.root)
            adapter = self._makeOne(plugin)
            self.assertEqual(len(adapter.listExportableItems()), 0)
            plugin.addGroup('group_id', 'title', 'description')
            self.assertEqual(len(adapter.listExportableItems()), 0)

        def test__getExportInfo_empty(self):
            plugin = self._makePlugin('empty').__of__(self.root)
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], None)
            self.assertEqual(len(info['groups']), 0)

        def test_export_empty(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('empty').__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len( context._wrote ), 1 )
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual( filename, 'plugins/empty.xml' )
            self._compareDOM( text, _EMPTY_ZODB_GROUPS )
            self.assertEqual( content_type, 'text/xml' )

        def test__getExportInfo_with_groups(self):

            plugin = self._makePlugin('with_groups').__of__(self.root)
            plugin.title = 'Plugin Title'

            for g in _ZODB_GROUP_INFO:
                plugin.addGroup(g['group_id'], g['title'], g['description'])
                for principal in g['principals']:
                    plugin.addPrincipalToGroup(principal, g['group_id'])

            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], 'Plugin Title')
            self.assertEqual(len(info['groups']), len(_ZODB_GROUP_INFO))

            for x, y in zip(info['groups'], _ZODB_GROUP_INFO):
                self.assertEqual(x, y)

        def test_export_with_groups(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('with_groups').__of__(self.root)
            plugin.title = 'Plugin Title'

            for g in _ZODB_GROUP_INFO:
                plugin.addGroup(g['group_id'], g['title'], g['description'])
                for principal in g['principals']:
                    plugin.addPrincipalToGroup(principal, g['group_id'])

            adapter = self._makeOne(plugin)
            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len(context._wrote), 1)
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual(filename, 'plugins/with_groups.xml')
            self._compareDOM(text, _FILLED_ZODB_GROUPS)
            self.assertEqual(content_type, 'text/xml')

        def test_import_empty(self):
            plugin = self._makePlugin('empty').__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin)
            context._files['plugins/empty.xml'] = _FILLED_ZODB_GROUPS
            self.assertEqual(plugin.title, None)

            adapter.import_(context, 'plugins', False)

            found = plugin.listGroupInfo()
            self.assertEqual(len(found), len(_ZODB_GROUP_INFO))
            self.assertEqual(plugin.title, 'Plugin Title')

            for finfo, ginfo in zip(found, _ZODB_GROUP_INFO):
                self.assertEqual(finfo['id'], ginfo['group_id'])
                self.assertEqual(finfo['title'], ginfo['title'])
                self.assertEqual(finfo['description'], ginfo['description'])
                for principal_id in ginfo['principals']:
                    groups = plugin._principal_groups[principal_id]
                    self.failUnless(ginfo['group_id'] in groups)

        def test_import_without_purge_leaves_existing_users(self):

            plugin = self._makePlugin('with_groups').__of__(self.root)
            plugin.title = 'Plugin Title'

            for g in _ZODB_GROUP_INFO:
                plugin.addGroup(g['group_id'], g['title'], g['description'])
                for principal in g['principals']:
                    plugin.addPrincipalToGroup(principal, g['group_id'])

            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, purge=False)
            context._files['plugins/with_groups.xml'] = _EMPTY_ZODB_GROUPS

            self.assertEqual(len(plugin._groups), len(_ZODB_GROUP_INFO))
            adapter.import_(context, 'plugins', False)
            self.assertEqual(len(plugin._groups), len(_ZODB_GROUP_INFO))
            self.assertEqual(plugin.title, None)

        def test_import_with_purge_wipes_existing_users(self):

            plugin = self._makePlugin('with_groups').__of__(self.root)
            plugin.title = 'Plugin Title'

            for g in _ZODB_GROUP_INFO:
                plugin.addGroup(g['group_id'], g['title'], g['description'])
                for principal in g['principals']:
                    plugin.addPrincipalToGroup(principal, g['group_id'])

            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, purge=True)
            context._files['plugins/with_groups.xml'] = _EMPTY_ZODB_GROUPS

            self.assertEqual(len(plugin._groups), len(_ZODB_GROUP_INFO))
            adapter.import_(context, 'plugins', False)
            self.assertEqual(len(plugin._groups), 0)
            self.assertEqual(plugin.title, None)

    class ZODBRoleManagerExportImportTests(_TestBase):

        def _getTargetClass(self):
            from Products.PluggableAuthService.plugins.exportimport \
                import ZODBRoleManagerExportImport
            return ZODBRoleManagerExportImport

        def _makePlugin(self, id, *args, **kw):
            from Products.PluggableAuthService.plugins.ZODBRoleManager \
                import ZODBRoleManager
            return ZODBRoleManager(id, *args, **kw)

        def test_listExportableItems(self):
            plugin = self._makePlugin('lEI').__of__(self.root)
            adapter = self._makeOne(plugin)

            self.assertEqual(len(adapter.listExportableItems()), 0)
            plugin.addRole('role_id', 'title', 'description')
            self.assertEqual(len(adapter.listExportableItems()), 0)

        def test__getExportInfo_empty(self):
            plugin = self._makePlugin('empty').__of__(self.root)
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], None)
            self.assertEqual(len(info['roles']), 0)

        def test_export_empty(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('empty').__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len( context._wrote ), 1 )
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual( filename, 'plugins/empty.xml' )
            self._compareDOM( text, _EMPTY_ZODB_ROLES )
            self.assertEqual( content_type, 'text/xml' )

        def test__getExportInfo_with_roles(self):

            plugin = self._makePlugin('with_roles').__of__(self.root)
            plugin.title = 'Plugin Title'

            for r in _ZODB_ROLE_INFO:
                plugin.addRole(r['role_id'], r['title'], r['description'])
                for principal in r['principals']:
                    plugin.assignRoleToPrincipal(r['role_id'], principal)

            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], 'Plugin Title')
            self.assertEqual(len(info['roles']), len(_ZODB_ROLE_INFO))

            for x, y in zip(info['roles'], _ZODB_ROLE_INFO):
                self.assertEqual(x, y)

        def test_export_with_roles(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('with_roles').__of__(self.root)
            plugin.title = 'Plugin Title'

            for r in _ZODB_ROLE_INFO:
                plugin.addRole(r['role_id'], r['title'], r['description'])
                for principal in r['principals']:
                    plugin.assignRoleToPrincipal(r['role_id'], principal)

            adapter = self._makeOne(plugin)
            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len(context._wrote), 1)
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual(filename, 'plugins/with_roles.xml')
            self._compareDOM(text, _FILLED_ZODB_ROLES)
            self.assertEqual(content_type, 'text/xml')

        def test_import_empty(self):
            plugin = self._makePlugin('empty').__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin)
            context._files['plugins/empty.xml'] = _FILLED_ZODB_ROLES
            self.assertEqual(plugin.title, None)

            adapter.import_(context, 'plugins', False)

            found = plugin.listRoleInfo()
            self.assertEqual(len(found), len(_ZODB_ROLE_INFO))
            self.assertEqual(plugin.title, 'Plugin Title')

            for finfo, rinfo in zip(found, _ZODB_ROLE_INFO):
                self.assertEqual(finfo['id'], rinfo['role_id'])
                self.assertEqual(finfo['title'], rinfo['title'])
                self.assertEqual(finfo['description'], rinfo['description'])
                for principal_id in rinfo['principals']:
                    roles = plugin._principal_roles[principal_id]
                    self.failUnless(rinfo['role_id'] in roles)

        def test_import_without_purge_leaves_existing_users(self):

            plugin = self._makePlugin('with_roles').__of__(self.root)
            plugin.title = 'Plugin Title'

            for r in _ZODB_ROLE_INFO:
                plugin.addRole(r['role_id'], r['title'], r['description'])
                for principal in r['principals']:
                    plugin.assignRoleToPrincipal(r['role_id'], principal)

            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, purge=False)
            context._files['plugins/with_roles.xml'] = _EMPTY_ZODB_ROLES

            self.assertEqual(len(plugin._roles), len(_ZODB_ROLE_INFO))
            adapter.import_(context, 'plugins', False)
            self.assertEqual(len(plugin._roles), len(_ZODB_ROLE_INFO))
            self.assertEqual(plugin.title, None)

        def test_import_with_purge_wipes_existing_users(self):

            plugin = self._makePlugin('with_roles').__of__(self.root)
            plugin.title = 'Plugin Title'

            for r in _ZODB_ROLE_INFO:
                plugin.addRole(r['role_id'], r['title'], r['description'])
                for principal in r['principals']:
                    plugin.assignRoleToPrincipal(r['role_id'], principal)

            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, purge=True)
            context._files['plugins/with_roles.xml'] = _EMPTY_ZODB_ROLES

            self.assertEqual(len(plugin._roles), len(_ZODB_ROLE_INFO))
            adapter.import_(context, 'plugins', False)
            self.assertEqual(len(plugin._roles), 0)
            self.assertEqual(plugin.title, None)

    class CookieAuthHelperExportImportTests(_TestBase):

        def _getTargetClass(self):
            from Products.PluggableAuthService.plugins.exportimport \
                import CookieAuthHelperExportImport
            return CookieAuthHelperExportImport

        def _makePlugin(self, id, *args, **kw):
            from Products.PluggableAuthService.plugins.CookieAuthHelper \
                import CookieAuthHelper
            return CookieAuthHelper(id, *args, **kw)

        def test_listExportableItems(self):
            plugin = self._makePlugin('lEI').__of__(self.root)
            adapter = self._makeOne(plugin)

            self.assertEqual(len(adapter.listExportableItems()), 0)
            plugin.cookie_name = 'COOKIE_NAME'
            plugin.login_path = 'LOGIN_PATH'
            self.assertEqual(len(adapter.listExportableItems()), 0)

        def test__getExportInfo_default(self):
            from Products.PluggableAuthService.plugins.CookieAuthHelper \
                import CookieAuthHelper
            plugin = self._makePlugin('default').__of__(self.root)
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], None)
            self.assertEqual(info['cookie_name'], CookieAuthHelper.cookie_name)
            self.assertEqual(info['login_path'], CookieAuthHelper.login_path)

        def test_export_default(self):
            _setUpDefaultTraversable()

            from Products.PluggableAuthService.plugins.CookieAuthHelper \
                import CookieAuthHelper as CAH
            plugin = self._makePlugin('default').__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual(len(context._wrote), 1)
            filename, text, content_type = context._wrote[0]
            self.assertEqual(filename, 'plugins/default.xml' )
            self._compareDOM(text,
                             _COOKIE_AUTH_TEMPLATE_NO_TITLE % (CAH.cookie_name,
                                                               CAH.login_path,
                                                              ))
            self.assertEqual( content_type, 'text/xml' )

        def test__getExportInfo_explicitly_set(self):
            TITLE = 'Plugin Title'
            COOKIE_NAME = 'COOKIE_NAME'
            LOGIN_PATH = 'LOGIN_PATH'
            plugin = self._makePlugin('explicit').__of__(self.root)
            plugin.title = TITLE
            plugin.cookie_name = COOKIE_NAME
            plugin.login_path = LOGIN_PATH
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], TITLE)
            self.assertEqual(info['cookie_name'], COOKIE_NAME)
            self.assertEqual(info['login_path'], LOGIN_PATH)

        def test_export_explicitly_set(self):
            _setUpDefaultTraversable()

            TITLE = 'Plugin Title'
            COOKIE_NAME = 'COOKIE_NAME'
            LOGIN_PATH = 'LOGIN_PATH'
            plugin = self._makePlugin('explicit').__of__(self.root)
            plugin.title = TITLE
            plugin.cookie_name = COOKIE_NAME
            plugin.login_path = LOGIN_PATH
            adapter = self._makeOne(plugin)

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual(len(context._wrote), 1)
            filename, text, content_type = context._wrote[0]
            self.assertEqual(filename, 'plugins/explicit.xml' )
            self._compareDOM(text,
                             _COOKIE_AUTH_TEMPLATE % (TITLE,
                                                      COOKIE_NAME,
                                                      LOGIN_PATH,
                                                     ))
            self.assertEqual( content_type, 'text/xml' )

        def test_import_with_title(self):
            TITLE = 'Plugin Title'
            COOKIE_NAME = 'COOKIE_NAME'
            LOGIN_PATH = 'LOGIN_PATH'
            plugin = self._makePlugin('with_title').__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin)
            context._files['plugins/with_title.xml'
                          ] = _COOKIE_AUTH_TEMPLATE % (TITLE,
                                                       COOKIE_NAME,
                                                       LOGIN_PATH,
                                                      )
            adapter.import_(context, 'plugins', False)

            self.assertEqual( plugin.title, TITLE )
            self.assertEqual( plugin.cookie_name, COOKIE_NAME )
            self.assertEqual( plugin.login_path, LOGIN_PATH )

        def test_import_no_title(self):
            TITLE = 'Plugin Title'
            COOKIE_NAME = 'COOKIE_NAME'
            LOGIN_PATH = 'LOGIN_PATH'
            plugin = self._makePlugin('no_title').__of__(self.root)
            plugin.title = TITLE
            plugin.cookie_name = COOKIE_NAME
            plugin.login_path = LOGIN_PATH
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin)
            context._files['plugins/no_title.xml'
                          ] = _COOKIE_AUTH_TEMPLATE_NO_TITLE % (COOKIE_NAME,
                                                                LOGIN_PATH,
                                                               )
            adapter.import_(context, 'plugins', False)

            self.assertEqual( plugin.title, None )
            self.assertEqual( plugin.cookie_name, COOKIE_NAME )
            self.assertEqual( plugin.login_path, LOGIN_PATH )

    class DomainAuthHelperExportImportTests(_TestBase):

        def _getTargetClass(self):
            from Products.PluggableAuthService.plugins.exportimport \
                import DomainAuthHelperExportImport
            return DomainAuthHelperExportImport

        def _makePlugin(self, id, *args, **kw):
            from Products.PluggableAuthService.plugins.DomainAuthHelper \
                import DomainAuthHelper
            return DomainAuthHelper(id, *args, **kw)

        def test_listExportableItems(self):
            plugin = self._makePlugin('lEI').__of__(self.root)
            adapter = self._makeOne(plugin)

            self.assertEqual(len(adapter.listExportableItems()), 0)
            plugin.cookie_name = 'COOKIE_NAME'
            plugin.login_path = 'LOGIN_PATH'
            plugin.manage_addMapping(user_id='user_id',
                                     match_type='equals',
                                     match_string='host.example.com',
                                     roles=['foo', 'bar'],
                                    )
            self.assertEqual(len(adapter.listExportableItems()), 0)

        def test__getExportInfo_empty(self):
            plugin = self._makePlugin('empty', None).__of__(self.root)
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], None)
            self.assertEqual(len(info['map']), 0)

        def test_export_empty(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('empty', None).__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len( context._wrote ), 1 )
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual( filename, 'plugins/empty.xml' )
            self._compareDOM( text, _EMPTY_DOMAIN_AUTH )
            self.assertEqual( content_type, 'text/xml' )

        def test__getExportInfo_with_map(self):
            TITLE = 'With Map'
            USER_ID = 'some_user_id'
            DOMAIN = 'host.example.com'
            ROLES = ['foo', 'bar']
            plugin = self._makePlugin('with_map', TITLE).__of__(self.root)
            adapter = self._makeOne(plugin)

            plugin.manage_addMapping(user_id=USER_ID,
                                     match_type='equals',
                                     match_string=DOMAIN,
                                     roles=ROLES,
                                    )

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], TITLE)
            user_map = info['map']
            self.assertEqual(len(user_map), 1)
            self.failUnless(USER_ID in user_map)
            match_list = user_map[USER_ID]
            self.assertEqual(len(match_list), 1)
            match = match_list[0]
            self.assertEqual(match['username'], USER_ID)
            self.assertEqual(match['match_type'], 'equals')
            self.assertEqual(match['match_string'], DOMAIN)
            self.assertEqual(match['roles'], ','.join(ROLES))

        def test_export_with_map(self):
            _setUpDefaultTraversable()

            TITLE = 'With Map'
            USER_ID = 'some_user_id'
            DOMAIN = 'host.example.com'
            ROLES = ['foo', 'bar']
            plugin = self._makePlugin('with_map', TITLE).__of__(self.root)
            adapter = self._makeOne(plugin)

            plugin.manage_addMapping(user_id=USER_ID,
                                     match_type='equals',
                                     match_string=DOMAIN,
                                     roles=ROLES,
                                    )

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual(len(context._wrote), 1)
            filename, text, content_type = context._wrote[0]
            self.assertEqual(filename, 'plugins/with_map.xml' )
            self._compareDOM(text,
                             _FILLED_DOMAIN_AUTH %
                              (TITLE,
                               USER_ID,
                               DOMAIN,
                               'equals',
                               ','.join(ROLES),
                               USER_ID,
                              ))
            self.assertEqual( content_type, 'text/xml' )

        def test_import_empty(self):
            from Products.PluggableAuthService.plugins.DomainAuthHelper \
                import EqualsFilter
            TITLE = 'With Map'
            USER_ID = 'some_user_id'
            DOMAIN = 'host.example.com'
            ROLES = ['foo', 'bar']
            plugin = self._makePlugin('empty').__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin)
            context._files['plugins/empty.xml'
                          ] = _FILLED_DOMAIN_AUTH % (TITLE,
                                                     USER_ID,
                                                     DOMAIN,
                                                     'equals',
                                                     ','.join(ROLES),
                                                     USER_ID,
                                                    )
            self.assertEqual(plugin.title, '')

            adapter.import_(context, 'plugins', False)

            self.assertEqual(len(plugin._domain_map), 1)
            self.assertEqual(plugin.title, TITLE)

            username, match_list = plugin._domain_map.items()[0]
            self.assertEqual(username, USER_ID)
            self.assertEqual(len(match_list), 1)
            match = match_list[0]
            self.assertEqual(match['username'], USER_ID)
            self.assertEqual(match['match_string'], DOMAIN)
            filter = match['match_filter']
            self.failUnless(isinstance(filter, EqualsFilter))
            self.assertEquals(filter.match_string, DOMAIN)
            self.assertEqual(match['match_type'], 'equals')
            self.assertEqual(len(match['roles']), len(ROLES))
            for role in ROLES:
                self.failUnless(role in match['roles'])

        def test_import_without_purge_leaves_existing_users(self):
            TITLE = 'With Map'
            USER_ID = 'some_user_id'
            DOMAIN = 'host.example.com'
            ROLES = ['foo', 'bar']
            plugin = self._makePlugin('with_map', TITLE).__of__(self.root)
            adapter = self._makeOne(plugin)

            plugin.manage_addMapping(user_id=USER_ID,
                                     match_type='equals',
                                     match_string=DOMAIN,
                                     roles=ROLES,
                                    )

            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, purge=False)
            context._files['plugins/with_map.xml'] = _EMPTY_DOMAIN_AUTH

            self.assertEqual(len(plugin._domain_map), 1)
            adapter.import_(context, 'plugins', False)
            self.assertEqual(len(plugin._domain_map), 1)
            self.assertEqual(plugin.title, None)

        def test_import_with_purge_wipes_existing_users(self):
            TITLE = 'With Map'
            USER_ID = 'some_user_id'
            DOMAIN = 'host.example.com'
            ROLES = ['foo', 'bar']
            plugin = self._makePlugin('with_map', TITLE).__of__(self.root)

            adapter = self._makeOne(plugin)

            plugin.manage_addMapping(user_id=USER_ID,
                                     match_type='equals',
                                     match_string=DOMAIN,
                                     roles=ROLES,
                                    )
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, purge=True)
            context._files['plugins/with_map.xml'] = _EMPTY_DOMAIN_AUTH

            self.assertEqual(len(plugin._domain_map), 1)
            adapter.import_(context, 'plugins', False)
            self.assertEqual(len(plugin._domain_map), 0)
            self.assertEqual(plugin.title, None)

    class TitleOnlyExportImportTests(_TestBase):

        def _getTargetClass(self):
            from Products.PluggableAuthService.plugins.exportimport \
                import TitleOnlyExportImport
            return TitleOnlyExportImport

        def _makePlugin(self, id, *args, **kw):
            from OFS.SimpleItem import SimpleItem

            class _Plugin(SimpleItem):
                title = None

                def __init__(self, id, title=None):
                    self._setId(id)
                    if title is not None:
                        self.title = title

            return _Plugin(id, *args, **kw)

        def test_listExportableItems(self):
            plugin = self._makePlugin('lEI').__of__(self.root)
            adapter = self._makeOne(plugin)

            self.assertEqual(len(adapter.listExportableItems()), 0)

        def test__getExportInfo_no_title(self):
            plugin = self._makePlugin('simple', None).__of__(self.root)
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], None)

        def test__getExportInfo_with_title(self):
            plugin = self._makePlugin('simple', None).__of__(self.root)
            plugin.title = 'Plugin Title'
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], 'Plugin Title')

        def test_export_no_title(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('no_title', None).__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len( context._wrote ), 1 )
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual( filename, 'plugins/no_title.xml' )
            self._compareDOM( text, _TITLE_ONLY_TEMPLATE_NO_TITLE )
            self.assertEqual( content_type, 'text/xml' )

        def test_export_with_title(self):
            _setUpDefaultTraversable()

            TITLE = 'Plugin Title'
            plugin = self._makePlugin('with_title', None).__of__(self.root)
            plugin.title = TITLE
            adapter = self._makeOne(plugin)

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len( context._wrote ), 1 )
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual( filename, 'plugins/with_title.xml' )
            self._compareDOM( text, _TITLE_ONLY_TEMPLATE % TITLE )
            self.assertEqual( content_type, 'text/xml' )

        def test_import_with_title(self):
            TITLE = 'Plugin Title'
            plugin = self._makePlugin('with_title').__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin)
            context._files['plugins/with_title.xml'
                          ] = _TITLE_ONLY_TEMPLATE % TITLE
            adapter.import_(context, 'plugins', False)

            self.assertEqual( plugin.title, TITLE )

        def test_import_no_title(self):
            TITLE = 'Plugin Title'
            plugin = self._makePlugin('no_title').__of__(self.root)
            plugin.title = TITLE
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin)
            context._files['plugins/no_title.xml'
                          ] = _TITLE_ONLY_TEMPLATE_NO_TITLE

            adapter.import_(context, 'plugins', False)

            self.assertEqual( plugin.title, None )

    class DelegatePathExportImportTests(_TestBase):

        def _getTargetClass(self):
            from Products.PluggableAuthService.plugins.exportimport \
                import DelegatePathExportImport
            return DelegatePathExportImport

        def _makePlugin(self, id, *args, **kw):
            from OFS.SimpleItem import SimpleItem

            class _Plugin(SimpleItem):
                title = None
                delegate = ''

                def __init__(self, id, title=None):
                    self._setId(id)
                    if title is not None:
                        self.title = title

            return _Plugin(id, *args, **kw)

        def test_listExportableItems(self):
            plugin = self._makePlugin('lEI').__of__(self.root)
            adapter = self._makeOne(plugin)

            self.assertEqual(len(adapter.listExportableItems()), 0)
            plugin.delegate = 'path/to/delegate'
            self.assertEqual(len(adapter.listExportableItems()), 0)

        def test__getExportInfo_default(self):
            plugin = self._makePlugin('default').__of__(self.root)
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], None)
            self.assertEqual(info['delegate'], '')

        def test_export_default(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('default').__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual(len(context._wrote), 1)
            filename, text, content_type = context._wrote[0]
            self.assertEqual(filename, 'plugins/default.xml' )
            self._compareDOM(text, _DELEGATE_PATH_TEMPLATE_NO_TITLE % "")
            self.assertEqual( content_type, 'text/xml' )

        def test__getExportInfo_explicitly_set(self):
            TITLE = 'Plugin Title'
            DELEGATE_PATH = 'path/to/delegate'
            plugin = self._makePlugin('explicit').__of__(self.root)
            plugin.title = TITLE
            plugin.delegate = DELEGATE_PATH
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], TITLE)
            self.assertEqual(info['delegate'], DELEGATE_PATH)

        def test_export_explicitly_set(self):
            _setUpDefaultTraversable()

            TITLE = 'Plugin Title'
            DELEGATE_PATH = 'path/to/delegate'
            plugin = self._makePlugin('explicit').__of__(self.root)
            plugin.title = TITLE
            plugin.delegate = DELEGATE_PATH
            adapter = self._makeOne(plugin)

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual(len(context._wrote), 1)
            filename, text, content_type = context._wrote[0]
            self.assertEqual(filename, 'plugins/explicit.xml' )
            self._compareDOM(text,
                             _DELEGATE_PATH_TEMPLATE % (TITLE,
                                                        DELEGATE_PATH,
                                                       ))
            self.assertEqual( content_type, 'text/xml' )

        def test_import_with_title(self):
            TITLE = 'Plugin Title'
            DELEGATE_PATH = 'path/to/delegate'
            plugin = self._makePlugin('with_title').__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin)
            context._files['plugins/with_title.xml'
                          ] = _DELEGATE_PATH_TEMPLATE % (TITLE,
                                                         DELEGATE_PATH,
                                                        )
            adapter.import_(context, 'plugins', False)

            self.assertEqual( plugin.title, TITLE )
            self.assertEqual( plugin.delegate, DELEGATE_PATH )

        def test_import_no_title(self):
            TITLE = 'Plugin Title'
            DELEGATE_PATH = 'path/to/delegate'
            plugin = self._makePlugin('no_title').__of__(self.root)
            plugin.title = TITLE
            plugin.delegate = DELEGATE_PATH
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin)
            context._files['plugins/no_title.xml'
                          ] = _DELEGATE_PATH_TEMPLATE_NO_TITLE % DELEGATE_PATH
            adapter.import_(context, 'plugins', False)

            self.assertEqual( plugin.title, None )
            self.assertEqual( plugin.delegate, DELEGATE_PATH )

    class DynamicGroupsPluginExportImportTests(_TestBase):

        def _getTargetClass(self):
            from Products.PluggableAuthService.plugins.exportimport \
                import DynamicGroupsPluginExportImport
            return DynamicGroupsPluginExportImport

        def _makePlugin(self, id, *args, **kw):
            from Products.PluggableAuthService.plugins.DynamicGroupsPlugin \
                import DynamicGroupsPlugin

            return DynamicGroupsPlugin(id, *args, **kw)

        def test_listExportableItems(self):
            plugin = self._makePlugin('lEI').__of__(self.root)
            adapter = self._makeOne(plugin)

            self.assertEqual(len(adapter.listExportableItems()), 0)
            plugin.addGroup('group_id', 'python:1', 'Group Title' )
            self.assertEqual(len(adapter.listExportableItems()), 0)

        def test__getExportInfo_empty(self):
            plugin = self._makePlugin('empty', None).__of__(self.root)
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], None)
            self.assertEqual(len(info['groups']), 0)

        def test_export_empty(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('empty', None).__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len( context._wrote ), 1 )
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual( filename, 'plugins/empty.xml' )
            self._compareDOM( text, _EMPTY_DYNAMIC_GROUPS )
            self.assertEqual( content_type, 'text/xml' )

        def test__getExportInfo_with_groups(self):

            plugin = self._makePlugin('with_groups').__of__(self.root)
            plugin.title = 'Plugin Title'

            for g in _DYNAMIC_GROUP_INFO:
                plugin.addGroup(g['group_id'],
                                g['predicate'],
                                g['title'],
                                g['description'],
                                g['active'],
                               )

            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], 'Plugin Title')
            self.assertEqual(len(info['groups']), len(_DYNAMIC_GROUP_INFO))

            for x, y in zip(info['groups'], _DYNAMIC_GROUP_INFO):
                self.assertEqual(x, y)

        def test_export_with_groups(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('with_groups').__of__(self.root)
            plugin.title = 'Plugin Title'

            for g in _DYNAMIC_GROUP_INFO:
                plugin.addGroup(g['group_id'],
                                g['predicate'],
                                g['title'],
                                g['description'],
                                g['active'],
                               )

            adapter = self._makeOne(plugin)
            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len(context._wrote), 1)
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual(filename, 'plugins/with_groups.xml')
            self._compareDOM(text, _FILLED_DYNAMIC_GROUPS)
            self.assertEqual(content_type, 'text/xml')

        def test_import_empty(self):
            plugin = self._makePlugin('empty', None).__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, encoding='ascii')
            context._files['plugins/empty.xml'] = _FILLED_DYNAMIC_GROUPS
            self.assertEqual(plugin.title, None)

            adapter.import_(context, 'plugins', False)

            found = plugin.listGroupInfo()
            self.assertEqual(len(found), len(_DYNAMIC_GROUP_INFO))
            self.assertEqual(plugin.title, 'Plugin Title')

            for finfo, ginfo in zip(found, _DYNAMIC_GROUP_INFO):
                self.assertEqual(finfo['id'], ginfo['group_id'])
                self.assertEqual(finfo['predicate'], ginfo['predicate'])
                self.assertEqual(finfo['title'], ginfo['title'])
                self.assertEqual(finfo['description'], ginfo['description'])
                self.assertEqual(finfo['active'], ginfo['active'])

        def test_import_without_purge_leaves_existing_users(self):

            plugin = self._makePlugin('with_groups').__of__(self.root)
            plugin.title = 'Plugin Title'

            for g in _DYNAMIC_GROUP_INFO:
                plugin.addGroup(g['group_id'],
                                g['predicate'],
                                g['title'],
                                g['description'],
                                g['active'],
                               )

            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, purge=False)
            context._files['plugins/with_groups.xml'] = _EMPTY_DYNAMIC_GROUPS

            self.assertEqual(len(plugin.listGroupIds()),
                             len(_DYNAMIC_GROUP_INFO))
            adapter.import_(context, 'plugins', False)
            self.assertEqual(len(plugin.listGroupIds()),
                             len(_DYNAMIC_GROUP_INFO))
            self.assertEqual(plugin.title, None)

        def test_import_with_purge_wipes_existing_users(self):

            plugin = self._makePlugin('with_groups').__of__(self.root)
            plugin.title = 'Plugin Title'

            for g in _DYNAMIC_GROUP_INFO:
                plugin.addGroup(g['group_id'],
                                g['predicate'],
                                g['title'],
                                g['description'],
                                g['active'],
                               )

            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, purge=True)
            context._files['plugins/with_groups.xml'] = _EMPTY_DYNAMIC_GROUPS

            self.assertEqual(len(plugin.listGroupIds()),
                             len(_DYNAMIC_GROUP_INFO))
            adapter.import_(context, 'plugins', False)
            self.assertEqual(len(plugin.listGroupIds()), 0)
            self.assertEqual(plugin.title, None)


    class ChallengeProtocolChooserExportImportTests(_TestBase):

        def _getTargetClass(self):
            from Products.PluggableAuthService.plugins.exportimport \
                import ChallengeProtocolChooserExportImport
            return ChallengeProtocolChooserExportImport

        def _makePlugin(self, id, *args, **kw):
            from \
              Products.PluggableAuthService.plugins.ChallengeProtocolChooser \
                import ChallengeProtocolChooser

            return ChallengeProtocolChooser(id, *args, **kw)

        def test_listExportableItems(self):
            plugin = self._makePlugin('lEI').__of__(self.root)
            adapter = self._makeOne(plugin)
            self.assertEqual(len(adapter.listExportableItems()), 0)
            plugin.manage_updateProtocolMapping({'WebDAV': ('http',)})
            self.assertEqual(len(adapter.listExportableItems()), 0)

        def test__getExportInfo_empty(self):
            from \
              Products.PluggableAuthService.plugins.ChallengeProtocolChooser \
                import listRequestTypesLabels
            labels = sorted(listRequestTypesLabels())
            plugin = self._makePlugin('empty', None).__of__(self.root)
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], None)
            self.assertEqual(len(info['request_types']), len(labels))
            for r_type, label in zip(info['request_types'], labels):
                self.assertEqual(r_type['label'], label)
                self.assertEqual(r_type['protocols'], [])

        def test_export_empty(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('empty', None).__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len( context._wrote ), 1 )
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual( filename, 'plugins/empty.xml' )
            self._compareDOM( text, _EMPTY_CHOOSER )
            self.assertEqual( content_type, 'text/xml' )

        def test__getExportInfo_non_empty(self):
            from \
              Products.PluggableAuthService.plugins.ChallengeProtocolChooser \
                import listRequestTypesLabels
            labels = sorted(listRequestTypesLabels())
            plugin = self._makePlugin('non_empty', None).__of__(self.root)
            plugin.title = 'TITLE'
            plugin.manage_updateProtocolMapping({'WebDAV': ('http',),
                                                 'FTP': ('http',),
                                                 'XML-RPC': ('http',),
                                                })
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], 'TITLE')
            self.assertEqual(len(info['request_types']), len(labels))
            for r_type, label in zip(info['request_types'], labels):
                self.assertEqual(r_type['label'], label)
                if label == 'Browser':
                    self.assertEqual(r_type['protocols'], [])
                else:
                    self.assertEqual(r_type['protocols'], ['http'])

        def test_export_with_map(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('with_map').__of__(self.root)
            plugin.title = 'Plugin Title'
            plugin.manage_updateProtocolMapping({'WebDAV': ('http', 'digest'),
                                                 'FTP': ('http',),
                                                 'XML-RPC': ('http', 'digest'),
                                                })


            adapter = self._makeOne(plugin)
            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len(context._wrote), 1)
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual(filename, 'plugins/with_map.xml')
            self._compareDOM(text, _FILLED_CHOOSER)
            self.assertEqual(content_type, 'text/xml')

        def test_import_empty(self):
            from \
              Products.PluggableAuthService.plugins.ChallengeProtocolChooser \
                import listRequestTypesLabels
            plugin = self._makePlugin('empty', None).__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, encoding='ascii')
            context._files['plugins/empty.xml'] = _EMPTY_CHOOSER
            self.assertEqual(plugin.title, None)

            adapter.import_(context, 'plugins', False)

            for label in  listRequestTypesLabels():
                protocols = sorted(plugin._map.get(label, ()))
                self.assertEqual(protocols, [])

        def test_import_non_empty(self):
            from \
              Products.PluggableAuthService.plugins.ChallengeProtocolChooser \
                import listRequestTypesLabels
            plugin = self._makePlugin('non_empty', None).__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, encoding='ascii')
            context._files['plugins/non_empty.xml'] = _FILLED_CHOOSER
            self.assertEqual(plugin.title, None)

            adapter.import_(context, 'plugins', False)

            for label in  listRequestTypesLabels():
                protocols = sorted(plugin._map.get(label, ()))
                if label == 'Browser':
                    self.assertEqual(protocols, [])
                elif label == 'FTP':
                    self.assertEqual(protocols, ['http'])
                else:
                    self.assertEqual(protocols, ['digest', 'http'])


    class NotCompetent_byRolesExportImportTests(_TestBase):

        def _getTargetClass(self):
            from Products.PluggableAuthService.plugins.exportimport \
                import NotCompetent_byRolesExportImport
            return NotCompetent_byRolesExportImport

        def _makePlugin(self, id, *args, **kw):
            from \
              Products.PluggableAuthService.plugins.NotCompetentHelper \
                import NotCompetent_byRoles

            return NotCompetent_byRoles(id, *args, **kw)

        def test_listExportableItems(self):
            plugin = self._makePlugin('ncwr').__of__(self.root)
            adapter = self._makeOne(plugin)
            self.assertEqual(len(adapter.listExportableItems()), 0)
            plugin.roles = ('Manager',)
            self.assertEqual(len(adapter.listExportableItems()), 0)

        def test__getExportInfo_empty(self):
            plugin = self._makePlugin('empty').__of__(self.root)
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], '')
            self.assertEqual(len(info['roles']), 0)

        def test_export_empty(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('empty', 'Plugin Title'
                                     ).__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyExportContext(plugin)
            adapter.export(context, 'plugins', False)

            self.assertEqual( len( context._wrote ), 1 )
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual( filename, 'plugins/empty.xml' )
            self._compareDOM( text, _EMPTY_NONCOMPETENT )
            self.assertEqual( content_type, 'text/xml' )

        def test__getExportInfo_non_empty(self):
            plugin = self._makePlugin('non_empty', 'TITLE').__of__(self.root)
            plugin.roles = ('Manager',)
            adapter = self._makeOne(plugin)

            info = adapter._getExportInfo()
            self.assertEqual(info['title'], 'TITLE')
            self.assertEqual(len(info['roles']), 1)
            self.assertEqual(info['roles'][0], 'Manager')

        def test_export_non_empty(self):
            _setUpDefaultTraversable()

            plugin = self._makePlugin('non_empty', 'Plugin Title'
                                     ).__of__(self.root)
            plugin.roles = ('Manager',)
            adapter = self._makeOne(plugin)
            context = DummyExportContext(plugin)

            adapter.export(context, 'plugins', False)

            self.assertEqual( len(context._wrote), 1)
            filename, text, content_type = context._wrote[ 0 ]
            self.assertEqual(filename, 'plugins/non_empty.xml')
            self._compareDOM(text, _FILLED_NONCOMPETENT)
            self.assertEqual(content_type, 'text/xml')

        def test_import_empty(self):
            plugin = self._makePlugin('empty', 'BEFORE').__of__(self.root)
            plugin.roles = ('Manager',)
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, encoding='ascii')
            context._files['plugins/empty.xml'] = _EMPTY_NONCOMPETENT

            adapter.import_(context, 'plugins', False)

            self.assertEqual(plugin.title, 'Plugin Title')
            self.assertEqual(plugin.roles, ())

        def test_import_non_empty(self):
            plugin = self._makePlugin('non_empty', 'BEFORE').__of__(self.root)
            adapter = self._makeOne(plugin)

            context = DummyImportContext(plugin, encoding='ascii')
            context._files['plugins/non_empty.xml'] = _FILLED_NONCOMPETENT

            adapter.import_(context, 'plugins', False)

            self.assertEqual(plugin.title, 'Plugin Title')
            self.assertEqual(plugin.roles, ('Manager',))


    def test_suite():
        return unittest.TestSuite((
            unittest.makeSuite(ZODBUserManagerExportImportTests),
            unittest.makeSuite(ZODBGroupManagerExportImportTests),
            unittest.makeSuite(ZODBRoleManagerExportImportTests),
            unittest.makeSuite(CookieAuthHelperExportImportTests),
            unittest.makeSuite(DomainAuthHelperExportImportTests),
            unittest.makeSuite(TitleOnlyExportImportTests),
            unittest.makeSuite(DelegatePathExportImportTests),
            unittest.makeSuite(DynamicGroupsPluginExportImportTests),
            unittest.makeSuite(ChallengeProtocolChooserExportImportTests),
            unittest.makeSuite(NotCompetent_byRolesExportImportTests),
        ))

_EMPTY_ZODB_USERS = """\
<?xml version="1.0" ?>
<zodb-users>
</zodb-users>
"""

_ZODB_USER_INFO = ({'user_id': 'user_1',
                    'login_name': 'user1@example.com',
                    'password': 'password1',
                   },
                   {'user_id': 'user_2',
                    'login_name': 'user2@example.com',
                    'password': 'password2',
                   },
                  )

_FILLED_ZODB_USERS = """\
<?xml version="1.0" ?>
<zodb-users title="Plugin Title">
<user user_id="user_1"
login_name="user1@example.com"
password_hash="%s" />
<user user_id="user_2"
login_name="user2@example.com"
password_hash="%s" />
</zodb-users>
"""

_EMPTY_ZODB_GROUPS = """\
<?xml version="1.0" ?>
<zodb-groups>
</zodb-groups>
"""

_ZODB_GROUP_INFO = ({'group_id': 'group_1',
                     'title': 'Group 1',
                     'description': 'First Group',
                     'principals': ('principal1', 'principal2'),
                    },
                    {'group_id': 'group_2',
                     'title': 'Group 2',
                     'description': 'Second Group',
                     'principals': ('principal1', 'principal3'),
                    },
                   )

_FILLED_ZODB_GROUPS = """\
<?xml version="1.0" ?>
<zodb-groups title="Plugin Title">
<group group_id="group_1" title="Group 1" description="First Group">
<principal principal_id="principal1" />
<principal principal_id="principal2" />
</group>
<group group_id="group_2" title="Group 2" description="Second Group">
<principal principal_id="principal1" />
<principal principal_id="principal3" />
</group>
</zodb-groups>
"""

_EMPTY_ZODB_ROLES = """\
<?xml version="1.0" ?>
<zodb-roles>
</zodb-roles>
"""

_ZODB_ROLE_INFO = ({'role_id': 'role_1',
                    'title': 'Role 1',
                    'description': 'First Role',
                    'principals': ('principal1', 'principal2'),
                   },
                   {'role_id': 'role_2',
                    'title': 'Role 2',
                    'description': 'Second Role',
                    'principals': ('principal1', 'principal3'),
                   },
                  )

_FILLED_ZODB_ROLES = """\
<?xml version="1.0" ?>
<zodb-roles title="Plugin Title">
<role role_id="role_1" title="Role 1" description="First Role">
<principal principal_id="principal1" />
<principal principal_id="principal2" />
</role>
<role role_id="role_2" title="Role 2" description="Second Role">
<principal principal_id="principal1" />
<principal principal_id="principal3" />
</role>
</zodb-roles>
"""

_COOKIE_AUTH_TEMPLATE_NO_TITLE = """\
<?xml version="1.0" ?>
<cookie-auth cookie_name="%s" login_path="%s" />
"""

_COOKIE_AUTH_TEMPLATE = """\
<?xml version="1.0" ?>
<cookie-auth title="%s" cookie_name="%s" login_path="%s" />
"""

_EMPTY_DOMAIN_AUTH = """\
<?xml version="1.0" ?>
<domain-auth>
</domain-auth>
"""

_FILLED_DOMAIN_AUTH = """\
<?xml version="1.0" ?>
<domain-auth title="%s">
 <user user_id="%s">
  <match match_string="%s" match_type="%s" roles="%s" username="%s"/>
 </user>
</domain-auth>
"""

_TITLE_ONLY_TEMPLATE_NO_TITLE = """\
<?xml version="1.0" ?>
<plug-in />
"""

_TITLE_ONLY_TEMPLATE = """\
<?xml version="1.0" ?>
<plug-in title="%s" />
"""

_DELEGATE_PATH_TEMPLATE_NO_TITLE = """\
<?xml version="1.0" ?>
<delegating-plugin delegate="%s" />
"""

_DELEGATE_PATH_TEMPLATE = """\
<?xml version="1.0" ?>
<delegating-plugin title="%s" delegate="%s" />
"""

_DYNAMIC_GROUP_INFO = ({'group_id': 'group_1',
                        'title': 'Group 1',
                        'predicate': 'python:1',
                        'description': 'First Group',
                        'active': True,
                       },
                       {'group_id': 'group_2',
                        'title': 'Group 2',
                        'predicate': 'python:0',
                        'description': 'Second Group',
                        'active': False,
                       },
                      )

_EMPTY_DYNAMIC_GROUPS = """\
<?xml version="1.0" ?>
<dynamic-groups>
</dynamic-groups>
"""

_FILLED_DYNAMIC_GROUPS = """\
<?xml version="1.0" ?>
<dynamic-groups title="Plugin Title">
<group
    group_id="group_1"
    predicate="python:1"
    title="Group 1"
    description="First Group"
    active="True"
    />
<group
    group_id="group_2"
    predicate="python:0"
    title="Group 2"
    description="Second Group"
    active="False"
    />
</dynamic-groups>
"""

_EMPTY_CHOOSER = """\
<?xml version="1.0" ?>
<challenge-protocol-chooser>
<mapping label="Browser" protocols="" />
<mapping label="FTP" protocols="" />
<mapping label="WebDAV" protocols="" />
<mapping label="XML-RPC" protocols="" />
</challenge-protocol-chooser>
"""

_FILLED_CHOOSER = """\
<?xml version="1.0" ?>
<challenge-protocol-chooser title="Plugin Title">
<mapping
    label="Browser"
    protocols=""
    />
<mapping
    label="FTP"
    protocols="http"
    />
<mapping
    label="WebDAV"
    protocols="digest,http"
    />
<mapping
    label="XML-RPC"
    protocols="digest,http"
    />
</challenge-protocol-chooser>
"""

_EMPTY_NONCOMPETENT = """\
<?xml version="1.0" ?>
<not-competent-by-roles title="Plugin Title">
</not-competent-by-roles>
"""

_FILLED_NONCOMPETENT = """\
<?xml version="1.0" ?>
<not-competent-by-roles title="Plugin Title">
<role>Manager</role>
</not-competent-by-roles>
"""



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest
from OFS.Folder import Folder
from OFS.SimpleItem import SimpleItem

from zope.interface import Interface
from zope.interface import providedBy

class IFaux( Interface ):

    def faux_method():
        pass

class IFauxTwo( Interface ):

    def two_method():
        pass

class DummyPluginRegistry( Folder ):

    def listPluginIds( self, interface ):
        return ()

    def _getInterfaceFromName( self, name ):
        if name == 'IFaux':
            return IFaux
        if name == 'IFauxTwo':
            return IFauxTwo

class ScriptablePluginTests( unittest.TestCase ):

    def _getTargetClass( self ):

        from Products.PluggableAuthService.plugins.ScriptablePlugin \
            import ScriptablePlugin

        return ScriptablePlugin

    def _makeOne( self, id='test', *args, **kw ):

        return self._getTargetClass()( id=id, *args, **kw )

    def test_empty( self ):

        scriptable_plugin = self._makeOne()
        self.failIf( IFaux in providedBy(scriptable_plugin) )
        self.failIf( IFauxTwo in providedBy(scriptable_plugin) )

    def test_withTwo( self ):

        parent = Folder()
        parent._setObject( 'plugins', DummyPluginRegistry() )

        scriptable_plugin = self._makeOne().__of__(parent)

        faux_method = SimpleItem( 'faux_method' )
        two_method = SimpleItem( 'two_method' )

        scriptable_plugin._setObject( 'faux_method', faux_method )
        scriptable_plugin._setObject( 'two_method', two_method )

        scriptable_plugin.manage_updateInterfaces( ['IFaux', 'IFauxTwo'] )

        self.failUnless( IFaux in providedBy(scriptable_plugin) )
        self.failUnless( IFauxTwo in providedBy(scriptable_plugin) )

    def test_withTwoOnlyOneWired( self ):

        parent = Folder()
        parent._setObject( 'plugins', DummyPluginRegistry() )

        scriptable_plugin = self._makeOne().__of__(parent)

        faux_method = SimpleItem( 'faux_method' )
        whatever = SimpleItem( 'whatever' )

        scriptable_plugin._setObject( 'faux_method', faux_method )
        scriptable_plugin._setObject( 'whatever', whatever )

        scriptable_plugin.manage_updateInterfaces( ['IFaux',] )

        self.failUnless( IFaux in providedBy(scriptable_plugin) )

    def test_withTwoMinusOne( self ):

        parent = Folder()
        parent._setObject( 'plugins', DummyPluginRegistry() )

        scriptable_plugin = self._makeOne().__of__(parent)

        faux_method = SimpleItem( 'faux_method' )
        two_method = SimpleItem( 'two_method' )

        scriptable_plugin._setObject( 'faux_method', faux_method )
        scriptable_plugin._setObject( 'two_method', two_method )

        scriptable_plugin.manage_updateInterfaces( ['IFaux', 'IFauxTwo'] )

        scriptable_plugin._delObject( 'two_method' )

        self.failUnless( IFaux in providedBy(scriptable_plugin) )
        self.failIf( IFauxTwo in providedBy(scriptable_plugin) )


if __name__ == '__main__':
    unittest.main()

def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite( ScriptablePluginTests ),
        ))



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Interfaces for PluggableAuthService

$Id$
"""

from zope.interface import Interface


class IExtractionPlugin( Interface ):

    """ Extracts login name and credentials from a request.
    """

    def extractCredentials( request ):

        """ request -> {...}

        o Return a mapping of any derived credentials.

        o Return an empty mapping to indicate that the plugin found no
          appropriate credentials.
        """

class ILoginPasswordExtractionPlugin( IExtractionPlugin ):

    """ Common-case derivative.
    """

    def extractCredentials( request ):

        """ request -> { 'login' : login
                       , 'password' : password
                       , k1 : v1
                       ,   ...
                       , kN : vN
                       } | {}

        o If credentials are found, the returned mapping will contain at
          least 'login' and 'password' keys, with the password in plaintext.

        o Return an empty mapping to indicate that the plugin found no
          appropriate credentials.
        """

class ILoginPasswordHostExtractionPlugin( ILoginPasswordExtractionPlugin ):

    """ Common-case derivative.
    """

    def extractCredentials( request ):

        """ request -> { 'login' : login
                       , 'password' : password
                       , 'remote_host' : remote_host
                       , 'remote_addr' : remote_addr
                       , k1 : v1
                       ,   ...
                       , kN : vN
                       } | {}

        o If credentials are found, the returned mapping will contain at
          least 'login', 'password', 'remote_host' and 'remote_addr' keys,
          with the password in plaintext.

        o Return an empty mapping to indicate that the plugin found no
          appropriate credentials.
        """

class IAuthenticationPlugin( Interface ):

    """ Map credentials to a user ID.
    """

    def authenticateCredentials( credentials ):

        """ credentials -> (userid, login)

        o 'credentials' will be a mapping, as returned by IExtractionPlugin.

        o Return a  tuple consisting of user ID (which may be different
          from the login name) and login

        o If the credentials cannot be authenticated, return None.
        """

class IChallengePlugin( Interface ):

    """ Initiate a challenge to the user to provide credentials.

        Challenge plugins have an attribute 'protocol' representing
        the protocol the plugin operates under, defaulting to None.

        Plugins operating under the same protocol will all be given an
        attempt to fire. The first plugin of a protocol group that
        successfully fires establishes the protocol of the overall
        challenge.
    """

    def challenge( request, response ):

        """ Assert via the response that credentials will be gathered.

        Takes a REQUEST object and a RESPONSE object.

        Returns True if it fired, False otherwise.

        Two common ways to initiate a challenge:

          - Add a 'WWW-Authenticate' header to the response object.

            NOTE: add, since the HTTP spec specifically allows for
            more than one challenge in a given response.

          - Cause the response object to redirect to another URL (a
            login form page, for instance)
        """

class ICredentialsUpdatePlugin( Interface ):

    """ Callback:  user has changed her password.

    This interface is not responsible for the actual password change,
    it is used after a successful password change event.
    """

    def updateCredentials( request, response, login, new_password ):

        """ Scribble as appropriate.
        """

class ICredentialsResetPlugin( Interface ):

    """ Callback:  user has logged out.
    """

    def resetCredentials( request, response ):

        """ Scribble as appropriate.
        """

class IUserAdderPlugin( Interface ):

    """ Create a new user record in a User Manager
    """

    def doAddUser( login, password ):

        """ Add a user record to a User Manager, with the given login
            and password.  It is up to the implementation to determine
            if the login is used as user id as well.

        o Return a Boolean indicating whether a user was added or not
        """

class IRoleAssignerPlugin( Interface ):

    """ Assign a role to an identified principal
    """

    def doAssignRoleToPrincipal( principal_id, role ):

        """ Create a principal/role association in a Role Manager

        o Return a Boolean indicating whether the role was assigned or not
        """

    def doRemoveRoleFromPrincipal( principal_id, role ):

        """ Remove a principal/role association from a Role Manager

        o Return a Boolean indicating whether the role was removed or not
        """

class IUserFactoryPlugin( Interface ):

    """ Create a new IPropertiedUser.
    """

    def createUser( user_id, name ):

        """ Return a user, if possible.

        o Return None to allow another plugin, or the default, to fire.
        """

class IAnonymousUserFactoryPlugin( Interface ):

    """ Create a new anonymous IPropertiedUser.
    """

    def createAnonymousUser():

        """ Return an anonymous user, if possible.

        o Return None to allow another plugin, or the default, to fire.
        """

class IPropertiesPlugin( Interface ):

    """ Return a property set for a user.
    """

    def getPropertiesForUser( user, request=None ):

        """ user -> {}

        o User will implement IPropertiedUser.

        o Plugin should return a dictionary or an object providing
          IPropertySheet.

        o Plugin may scribble on the user, if needed (but must still
          return a mapping, even if empty).

        o May assign properties based on values in the REQUEST object, if
          present
        """

class IGroupsPlugin( Interface ):

    """ Determine the groups to which a user belongs.
    """

    def getGroupsForPrincipal( principal, request=None ):

        """ principal -> ( group_1, ... group_N )

        o Return a sequence of group names to which the principal
          (either a user or another group) belongs.

        o May assign groups based on values in the REQUEST object, if present
        """

class IRolesPlugin( Interface ):

    """ Determine the (global) roles which a user has.
    """

    def getRolesForPrincipal( principal, request=None ):

        """ principal -> ( role_1, ... role_N )

        o Return a sequence of role names which the principal has.

        o May assign roles based on values in the REQUEST object, if present.
        """

class IUpdatePlugin( Interface ):

    """ Allow the user or the application to update the user's properties.
    """

    def updateUserInfo( user, set_id, set_info ):

        """ Update backing store for 'set_id' using 'set_info'.
        """

class IValidationPlugin( Interface ):

    """ Specify allowable values for user properties.

    o E.g., constrain minimum password length, allowed characters, etc.

    o Operate on entire property sets, not individual properties.
    """

    def validateUserInfo( user, set_id, set_info ):

        """ -> ( error_info_1, ... error_info_N )

        o Returned values are dictionaries, containing at least keys:

          'id' -- the ID of the property, or None if the error is not
                  specific to one property.

          'error' -- the message string, suitable for display to the user.
        """

class IUserEnumerationPlugin( Interface ):

    """ Allow querying users by ID, and searching for users.

    o XXX:  can these be done by a single plugin?
    """

    def enumerateUsers( id=None
                      , login=None
                      , exact_match=False
                      , sort_by=None
                      , max_results=None
                      , **kw
                      ):

        """ -> ( user_info_1, ... user_info_N )

        o Return mappings for users matching the given criteria.

        o 'id' or 'login', in combination with 'exact_match' true, will
          return at most one mapping per supplied ID ('id' and 'login'
          may be sequences).

        o If 'exact_match' is False, then 'id' and / or login may be
          treated by the plugin as "contains" searches (more complicated
          searches may be supported by some plugins using other keyword
          arguments).

        o If 'sort_by' is passed, the results will be sorted accordingly.
          known valid values are 'id' and 'login' (some plugins may support
          others).

        o If 'max_results' is specified, it must be a positive integer,
          limiting the number of returned mappings.  If unspecified, the
          plugin should return mappings for all users satisfying the criteria.

        o Minimal keys in the returned mappings:

          'id' -- (required) the user ID, which may be different than
                  the login name

          'login' -- (required) the login name

          'pluginid' -- (required) the plugin ID (as returned by getId())

          'editurl' -- (optional) the URL to a page for updating the
                       mapping's user

        o Plugin *must* ignore unknown criteria.

        o Plugin may raise ValueError for invalid criteria.

        o Insufficiently-specified criteria may have catastrophic
          scaling issues for some implementations.
        """

    def updateUser( user_id, login_name ):
        """ Update the login name of the user with id user_id.

        The plugin must return True (or any truth value) to indicate a
        successful update, also when no update was needed.

        When updating a login name makes no sense for a plugin (most
        likely because it does not actually store login names) and it
        does not do anything, it must return None or False.
        """

    def updateEveryLoginName(quit_on_first_error=True):
        """Update login names of all users to their canonical value.

        This should be done after changing the login_transform
        property of PAS.

        You can set quit_on_first_error to False to report all errors
        before quitting with an error.  This can be useful if you want
        to know how many problems there are, if any.
        """

class IGroupEnumerationPlugin( Interface ):

    """ Allow querying groups by ID, and searching for groups.

    o XXX:  can these be done by a single plugin?
    """

    def enumerateGroups( id=None
                       , exact_match=False
                       , sort_by=None
                       , max_results=None
                       , **kw
                       ):

        """ -> ( group_info_1, ... group_info_N )

        o Return mappings for groups matching the given criteria.

        o 'id' in combination with 'exact_match' true, will
          return at most one mapping per supplied ID ('id'
          may be a sequence).

        o If 'exact_match' is False, then 'id' may be treated by
          the plugin as "contains" searches (more complicated searches
          may be supported by some plugins using other keyword arguments).

        o If 'sort_by' is passed, the results will be sorted accordingly.
          known valid values are 'id' (some plugins may support others).

        o If 'max_results' is specified, it must be a positive integer,
          limiting the number of returned mappings.  If unspecified, the
          plugin should return mappings for all groups satisfying the
          criteria.

        o Minimal keys in the returned mappings:

          'id' -- (required) the group ID

          'pluginid' -- (required) the plugin ID (as returned by getId())

          'properties_url' -- (optional) the URL to a page for updating the
                              group's properties.

          'members_url' -- (optional) the URL to a page for updating the
                           principals who belong to the group.

        o Plugin *must* ignore unknown criteria.

        o Plugin may raise ValueError for invalid critera.

        o Insufficiently-specified criteria may have catastrophic
          scaling issues for some implementations.
        """

class IRoleEnumerationPlugin( Interface ):

    """ Allow querying roles by ID, and searching for roles.
    """
    def enumerateRoles( id=None
                      , exact_match=False
                      , sort_by=None
                      , max_results=None
                      , **kw
                      ):

        """ -> ( role_info_1, ... role_info_N )

        o Return mappings for roles matching the given criteria.

        o 'id' in combination with 'exact_match' true, will
          return at most one mapping per supplied ID ('id'
          may be a sequence).

        o If 'exact_match' is False, then 'id' may be treated by
          the plugin as "contains" searches (more complicated searches
          may be supported by some plugins using other keyword arguments).

        o If 'sort_by' is passed, the results will be sorted accordingly.
          known valid values are 'id' (some plugins may support others).

        o If 'max_results' is specified, it must be a positive integer,
          limiting the number of returned mappings.  If unspecified, the
          plugin should return mappings for all roles satisfying the
          criteria.

        o Minimal keys in the returned mappings:

          'id' -- (required) the role ID

          'pluginid' -- (required) the plugin ID (as returned by getId())

          'properties_url' -- (optional) the URL to a page for updating the
                              role's properties.

          'members_url' -- (optional) the URL to a page for updating the
                           principals to whom the role is assigned.

        o Plugin *must* ignore unknown criteria.

        o Plugin may raise ValueError for invalid critera.

        o Insufficiently-specified criteria may have catastrophic
          scaling issues for some implementations.
        """

class IRequestTypeSniffer( Interface ):

    """ Given a request, detects the request type for later use by other plugins.
    """
    def sniffRequestType( request ):
        """ Return a interface identifying what kind the request is.
        """

class IChallengeProtocolChooser( Interface ):

    """ Choose a proper set of protocols to be used for challenging
    the client given a request.
    """
    def chooseProtocols( request ):
        """ -> ( protocol_1, ... protocol_N) | None

        o If a set of protocols is returned, the first plugin with a
            protocol that is in the set will define the protocol to be
            used for the current request.

        o If None is returned, the 'first found protocol' wins.

        o Once the protocol is decided, all challenge plugins for that
            protocol will be executed.
        """
#
#   XXX:  Do we need a LocalRoleAlgorithm plugin type?  E.g., base_cms
#         has two different algorithms, based on whether or not the
#         context object implements IPlacelessSecurity.
#


class INotCompetentPlugin( Interface ):

    """check whether this user folder is not competent to authenticate.

    Never used for a top level user folder;
    primarily used to prevent shaddowing of authentications by higher level
    user folders.
    """

    def isNotCompetentToAuthenticate(request):

        """return true if this user folder should not authenticate *request*.
        """
